/***************************************************************
* L&L - Labyrinths & Legends
* Copyright (c) 1993-2008 YOSHIMURA Tomohiko All rights reserved.
* 
* Created by BowKenKen
*   URL: https://sourceforge.jp/projects/lnl/
* 
* License is GPL
* 
* ܥץϥե꡼եȥǤ
* ʤϡ Free Software Foundation ɽ
*  GNU ̸ͭѵΡ֥С󣲡
* ϤʹߤγƥС椫餤줫򤷡
* ΥС˽äܥץ
* ۤޤѹ뤳ȤǤޤ
* 
* ܥץͭѤȤϻפޤۤˤäƤϡ
* ԾڤŪŬˤĤƤΰۤݾڤޤ,
* ʤݾڤԤʤޤ
* ܺ٤ˤĤƤ GNU ̸ͭѵɤߤ
* 
* ʤϡܥץȰ GNU ̸ͭѵ
* μ̤äƤϤǤǤʤϡ
*   Free Software Foundation, Inc.,
*   59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
* ؼ񤤤Ƥ
* 
* $Id: PcgDun.cpp,v 1.62 2008/04/03 10:26:45 bowkenken Exp $
***************************************************************/

#include	"gmain.h"
#include	"misc.h"
#include	"turn.h"
#include	"dun.h"
#include	"town.h"
//#include	"item.h"
//#include	"spell.h"
#include	"chr.h"
#include	"party.h"
#include	"mnstr.h"
//#include	"pet.h"
//#include	"fight.h"
//#include	"fx.h"
#include	"trap.h"
#include	"draw.h"
//#include	"curs.h"
#include	"menu.h"
#include	"amenu.h"
//#include	"request.h"
#include	"gfile.h"
#include	"msg.h"
//#include	"ver.h"
#include	"gmain-prot.h"
#include	"misc-prot.h"
#include	"turn-prot.h"
#include	"dun-prot.h"
//#include	"town-prot.h"
//#include	"item-prot.h"
//#include	"spell-prot.h"
#include	"chr-prot.h"
#include	"party-prot.h"
#include	"mnstr-prot.h"
//#include	"pet-prot.h"
//#include	"fight-prot.h"
//#include	"fx-prot.h"
#include	"trap-prot.h"
#include	"draw-prot.h"
//#include	"curs-prot.h"
//#include	"menu-prot.h"
#include	"mmenu-prot.h"
//#include	"imenu-prot.h"
//#include	"tmenu-prot.h"
//#include	"amenu-prot.h"
//#include	"request-prot.h"
#include	"gfile-prot.h"
#include	"msg-prot.h"

#include "PcgDun.h"
#include "FileList.h"
#include "GlobalVar.h"

#ifdef D_WS
# include <WSCform.h>
# include <WSCvscrBar.h>
#endif // D_WS

#ifdef D_GTK
# include <gtk/gtk.h>
# include <gdk-pixbuf/gdk-pixbuf.h>
#endif // D_GTK

#ifdef D_MFC
# include "xlnl-dows/xlnl-dows.h"
# include "xlnl-dows/MainFrm.h"
#endif // D_GTK

#ifdef D_WS
extern WSCform *MapForm;
extern WSCvscrBar *MapScrollX;
extern WSCvscrBar *MapScrollY;
#endif // D_WS

#include "PcgDunTab.h"

////////////////////////////////////////////////////////////////

// ѥΥޥå׾襵Ϥ߽ФƤɤ;
#define MAP_DRAW_ADD_R	1

// Ĥʸ TextOut() Ȥä褹뤫
#define FLAG_USE_TEXT_OUT	0

// 롦եȡ˹碌뤫
#define FLAG_FONT_TO_TEXT_SIZE	0

// ǥեȤΥǥ쥯ȥ
#define STR_DEFAULT_GRAPH_DIR_NAME	"xlnl-" VERSION "/"

// ̥ǥ쥯ȥ
#define STR_FIX_GRAPH_DIR_NAME	".xlnl/"
// ȥΥǥ쥯ȥ
#define STR_TITLE_DIR_NAME	"title/"

#define TITLE_FONT_POINT	32
#define TITLE_FONT_DOT	32

////////////////////////////////////////////////////////////////
// ѥ󡦥ǥåΥơ֥
////////////////////////////////////////////////////////////////

const long gBitIdxWallN = 1 << 0;
const long gBitIdxWallS = 1 << 1;
const long gBitIdxWallE = 1 << 2;
const long gBitIdxWallW = 1 << 3;

pos_t gIdxWall[2 * 2 * 2 * 2] = {
	{ +3, +3 }, //....
	{ +3, +2 }, //...N
	{ +3, +0 }, //..S.
	{ +3, +1 }, //..SN

	{ +0, +3 }, //.E.
	{ +0, +2 }, //.E.N
	{ +0, +0 }, //.ES.
	{ +0, +1 }, //.ESN

	{ +2, +3 }, //W...
	{ +2, +2 }, //W..N
	{ +2, +0 }, //W.S.
	{ +2, +1 }, //W.SN

	{ +1, +3 }, //WE..
	{ +1, +2 }, //WE.N
	{ +1, +0 }, //WES.
	{ +1, +1 }, //WESN
};

const long gBitIdxWallNE = 1 << 0;
const long gBitIdxWallNW = 1 << 1;
const long gBitIdxWallSE = 1 << 2;
const long gBitIdxWallSW = 1 << 3;

pos_t gIdxWallDiagonal[2 * 2 * 2 * 2] = {
	{ +1, +1 }, //.. .. .. ..
	{ +1, +1 }, //.. .. .. NE
	{ +1, +1 }, //.. .. NW ..
	{ +1, +1 }, //.. .. NW NE

	{ +1, +1 }, //.. SE .. ..
	{ +1, +1 }, //.. SE .. NE
	{ +1, +1 }, //.. SE NW ..
	{ +1, +4 }, //.. SE NW NE

	{ +1, +1 }, //SW .. .. ..
	{ +1, +1 }, //SW .. .. NE
	{ +1, +1 }, //SW .. NW ..
	{ +0, +4 }, //SW .. NW NE

	{ +1, +1 }, //SW SE .. ..
	{ +3, +4 }, //SW SE .. NE
	{ +2, +4 }, //SW SE NW ..
	{ +1, +1 }, //SW SE NW NE
};

////////////////////////////////////////////////////////////////

static const char *gStrTitle[3] = {
	"Labyrinths",
	"   and    ",
	" Legends  ",
};

typedef enum {
	DRAW_TITLE_KIND_0,
	DRAW_TITLE_KIND_1,
	DRAW_TITLE_KIND_MAX_N,
} drawTitleKindType;

////////////////////////////////////////////////////////////////

#ifdef D_GTK
static GdkFont *gSelFont;

static void handle_sel_font_destroy(
	GtkWidget *widget, gpointer user_data
);
static void handle_sel_font_ok(
	GtkWidget *widget, gpointer *data
);
#endif // D_GTK

////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////

PcgDun::PcgDun()
{
	pWBuf = NULL;

	nTileSizeX = 32;
	nTileSizeY = 32;
	nTileSizeRate = _100_PERCENT;

	nPtnDirMbr = N_PTN_DIR_CHR_1;
	nPtnDirMnstr = N_PTN_DIR_CHR_1;

	nTextModeDefaultPoint = 16;

	pStyle = NULL;
	nTextSizeX = 16;
	nTextSizeY = 16;
	nIntervalWidth = 1;
	nIntervalHeight = 1;

	pTitleStyle = NULL;

	bFlgEnaDrawTurn = false;

#ifdef D_MFC
	pDc = NULL;
	pSignboardFont = NULL;
#endif // D_MFC

	resetUpdateFlg();
}

////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////

PcgDun::~PcgDun()
{
	if( pWBuf != NULL ){
		delete pWBuf;
		pWBuf = NULL;
	}

#ifdef D_MFC
	if( pSignboardFont != NULL ){
		delete pSignboardFont;
		pSignboardFont = NULL;
	}

	if( pTitleFont != NULL ){
		delete pTitleFont;
		pTitleFont = NULL;
	}

	if( pDc != NULL ){
		delete pDc;
		pDc = NULL;
	}
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ȥν
////////////////////////////////////////////////////////////////

void PcgDun::initTitle()
{
	// ȥθ

	WSCstring dir = STR_FIX_GRAPH_DIR_NAME;
	WSCstring ext = STR_GRAPH_FILE_EXT;

	FileList::setStrDirSelGraph( dir );
	FileList ls;

	// ȥ򸡺

	{
		nTitleN = 0;

		ls.reset( STR_TITLE_DIR_NAME, ext );
		long j = 0;
		for( j = 0; j < LOOP_MAX_1000; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			if( randm( j + 1 ) == 0 )
				nTitleN = j;
		}
		nTitleN = j;
	}

	// ȥ

	if( nTitleN > 0 )
		nTitleRandmSelN = randm( nTitleN );
	else
		nTitleRandmSelN = -1;

	// ȥɤ߹

	if( nTitleRandmSelN > -1 ){
		ls.reset( STR_TITLE_DIR_NAME, ext );
		long j = 0;
		for( j = 0; j < LOOP_MAX_1000; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			if( j == nTitleRandmSelN ){
				pcgTitle.init( path );
				break;
			}
		}
	}

#ifdef D_GTK
	// ǥեȤΥ

	if( pTitleStyle == NULL )
		pTitleStyle = gtk_style_copy( gtk_widget_get_default_style() );
	if( pTitleStyle->fg_gc[GTK_STATE_NORMAL] == NULL ){
		pTitleStyle->fg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}
	if( pTitleStyle->bg_gc[GTK_STATE_NORMAL] == NULL ){
		pTitleStyle->bg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}

	// եȤ

	WSCstring sFontPoint = TITLE_FONT_POINT * 10;
	WSCstring sFontName = "";
	sFontName += "-*-*-*-i-normal--*-";
	sFontName += sFontPoint;
	sFontName += "-*,-*";

	gtk_style_set_font( pTitleStyle, gdk_fontset_load( sFontName ) );
#endif // D_GTK

#ifdef D_MFC
	// ǥեȤΥ

	if( pTitleStyle == NULL )
		pTitleStyle = (void *)1;

	// եȤ

	LONG h = TITLE_FONT_DOT;
	mTitleFontDesc.lfHeight = h;
	mTitleFontDesc.lfWidth = 0;
	mTitleFontDesc.lfEscapement = 0;
	mTitleFontDesc.lfOrientation = 0;
	mTitleFontDesc.lfWeight = FW_NORMAL;
	mTitleFontDesc.lfItalic = FALSE;
	mTitleFontDesc.lfUnderline = FALSE;
	mTitleFontDesc.lfStrikeOut = FALSE;
	mTitleFontDesc.lfCharSet = SHIFTJIS_CHARSET;
	mTitleFontDesc.lfOutPrecision
			= OUT_DEFAULT_PRECIS;
	mTitleFontDesc.lfClipPrecision
			= CLIP_DEFAULT_PRECIS;
	mTitleFontDesc.lfQuality = DEFAULT_QUALITY;
	mTitleFontDesc.lfPitchAndFamily
			= (FIXED_PITCH | FF_MODERN);
	strcpy( mTitleFontDesc.lfFaceName, "Pica" );

	pTitleFont = new CFont;
	pTitleFont->CreateFontIndirect(
			&mTitleFontDesc );

	mD3Title.Create( &mTitleFontDesc );
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// 
// GraphConf *cnf : եå
////////////////////////////////////////////////////////////////

void PcgDun::init( GraphConf *cnf )
{
	if( cnf == NULL )
		exit_game( EXIT_FAILURE );

	initGraphConf( cnf );

	resetSignboardFont();

	if( g_flg_text_mode ){
		resetUpdateFlg();
		return;
	}

	// ѥ󸡺

	WSCstring path = cnf->getDir();
	long w = path.getWords( "/" );
	WSCstring dir = path.getWord( w - 1, "/" );

	WSCstring ext = STR_GRAPH_FILE_EXT;

	FileList::setStrDirSelGraph( dir );
	FileList ls;

	// 롦СĴ

#ifdef D_MFC
	CXlnlView *view = theApp.m_pMainFrm->GetView();
	view->ResetScroll();
#endif // D_MFC

	// ̤õ

	for( long i = 0; i < PCG_NULL_MAX_N; i++ ){
		nNullTownN[i] = 0;

		if( gPcgTabNullTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabNullTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aNullTown[i][j].init( path );
		}
		nNullTownN[i] = j;
	}
	for( long i = 0; i < PCG_NULL_MAX_N; i++ ){
		nNullDunN[i] = 0;

		if( gPcgTabNullDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabNullDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aNullDun[i][j].init( path );
		}
		nNullDunN[i] = j;
	}

	// 

	for( long i = 0; i < PCG_FLOOR_MAX_N; i++ ){
		nFloorTownN[i] = 0;

		if( gPcgTabFloorTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabFloorTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aFloorTown[i][j].init( path );
		}
		nFloorTownN[i] = j;
	}
	for( long i = 0; i < PCG_FLOOR_MAX_N; i++ ){
		nFloorDunN[i] = 0;

		if( gPcgTabFloorDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabFloorDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aFloorDun[i][j].init( path );
		}
		nFloorDunN[i] = j;
	}

	// 

	for( long i = 0; i < PCG_WALL_MAX_N; i++ ){
		nWallTownN[i] = 0;

		if( gPcgTabWallTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabWallTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aWallTown[i][j].init( path );
		}
		nWallTownN[i] = j;
	}
	for( long i = 0; i < PCG_WALL_MAX_N; i++ ){
		nWallDunN[i] = 0;

		if( gPcgTabWallDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabWallDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aWallDun[i][j].init( path );
		}
		nWallDunN[i] = j;
	}

	// 

	for( long i = 0; i < PCG_SIGNBOARD_MAX_N; i++ ){
		nSignboardN[i] = 0;

		if( gPcgTabSignboard[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabSignboard[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aSignboard[i][j].init( path );
		}
		nSignboardN[i] = j;
	}

	// 곬

	for( long i = 0; i < PCG_STAIRS_MAX_N; i++ ){
		nStairsUpTownN[i] = 0;

		if( gPcgTabStairsUpTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabStairsUpTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aStairsUpTown[i][j].init( path );
		}
		nStairsUpTownN[i] = j;
	}
	for( long i = 0; i < PCG_STAIRS_MAX_N; i++ ){
		nStairsUpDunN[i] = 0;

		if( gPcgTabStairsUpDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabStairsUpDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aStairsUpDun[i][j].init( path );
		}
		nStairsUpDunN[i] = j;
	}

	// 곬

	for( long i = 0; i < PCG_STAIRS_MAX_N; i++ ){
		nStairsDownTownN[i] = 0;

		if( gPcgTabStairsDownTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabStairsDownTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aStairsDownTown[i][j].init( path );
		}
		nStairsDownTownN[i] = j;
	}
	for( long i = 0; i < PCG_STAIRS_MAX_N; i++ ){
		nStairsDownDunN[i] = 0;

		if( gPcgTabStairsDownDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabStairsDownDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aStairsDownDun[i][j].init( path );
		}
		nStairsDownDunN[i] = j;
	}

	// Ĥޤäɥ

	for( long i = 0; i < PCG_DOOR_MAX_N; i++ ){
		nDoorCloseTownN[i] = 0;

		if( gPcgTabDoorCloseTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabDoorCloseTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aDoorCloseTown[i][j].init( path );
		}
		nDoorCloseTownN[i] = j;
	}
	for( long i = 0; i < PCG_DOOR_MAX_N; i++ ){
		nDoorCloseDunN[i] = 0;

		if( gPcgTabDoorCloseDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabDoorCloseDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aDoorCloseDun[i][j].init( path );
		}
		nDoorCloseDunN[i] = j;
	}

	// ɥ

	for( long i = 0; i < PCG_DOOR_MAX_N; i++ ){
		nDoorOpenTownN[i] = 0;

		if( gPcgTabDoorOpenTown[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabDoorOpenTown[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aDoorOpenTown[i][j].init( path );
		}
		nDoorOpenTownN[i] = j;
	}
	for( long i = 0; i < PCG_DOOR_MAX_N; i++ ){
		nDoorOpenDunN[i] = 0;

		if( gPcgTabDoorOpenDun[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabDoorOpenDun[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aDoorOpenDun[i][j].init( path );
		}
		nDoorOpenDunN[i] = j;
	}

	// 

	for( long i = 0; i < PCG_TRAP_MAX_N; i++ ){
		nTrapN[i] = 0;

		if( gPcgTabTrap[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabTrap[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aTrap[i][j].init( path );
		}
		nTrapN[i] = j;
	}

	// 

	for( long i = 0; i < PCG_CRSR_MAX_N; i++ ){
		nCrsrN[i] = 0;

		if( gPcgTabCrsr[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabCrsr[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aCrsr[i][j].init( path );
		}
		nCrsrN[i] = j;
	}

	// 󥫡

	for( long i = 0; i < PCG_SQUARE_MAX_N; i++ ){
		nSquareN[i] = 0;

		if( gPcgTabSquare[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabSquare[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aSquare[i][j].init( path );
		}
		nSquareN[i] = j;
	}

	// ƥ

	for( long i = 0; i < PCG_ITEM_MAX_N; i++ ){
		nItemN[i] = 0;

		if( gPcgTabItem[i].nRatio <= -1 )
			break;

		ls.reset( gPcgTabItem[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aItem[i][j].init( path );
		}
		nItemN[i] = j;
	}

	// С

	mbr_t **ls_mbr = get_ls_mbr();

	for( long i = 0; i < LS_MBR_MAX_N; i++ ){
		ls.reset( gPcgTabMbrDflt.sDir, ext );
		WSCstring path = ls.jointDir( ls.getBaseDir(),
				gPcgTabMbrDflt.sDir );

		bool flagExist = true;
		if( ls_mbr[i] == NULL )
			flagExist = false;
		if( chk_flg( ls_mbr[i]->stat, FLG_STAT_NOT_EXIST ) )
			flagExist = false;
		if( ls_mbr[i]->graph_name[0] == '\0' )
			flagExist = false;

		if( flagExist ){
			path = ls.jointDir( path, ls_mbr[i]->graph_name );

			if( chk_exist_file( path ) ){
				aMbr[i].init( path );
				continue;
			}
		}

		ls.reset( gPcgTabMbrNull.sDir, ext );
		WSCstring file = ls.next();
		if( file.getChars() <= 0 )
			continue;
		aMbr[i].init( file );
	}

	// 󥹥

	for( long i = 0; i < MNSTR_KIND_MAX_N; i++ ){
		nMnstrN[i] = 0;
	}
	for( long i = 0; i < PCG_MNSTR_MAX_N; i++ ){
		if( gPcgTabMnstr[i].nRatio <= -1 )
			break;

		mnstr_kind_t kind = gPcgTabMnstr[i].nMnstrKind;
		if( kind == MNSTR_KIND_NULL )
			break;

		nMnstrN[kind] = 0;
		ls.reset( gPcgTabMnstr[i].sDir, ext );
		long j = 0;
		for( j = 0; j < PCG_LS_MAX_N; j++ ){
			WSCstring path = ls.next();
			if( path.getChars() <= 0 )
				break;
			aMnstr[kind][j].init( path );

			aMnstr[kind][j].dis();
		}
		nMnstrN[kind] = j;
	}
}

////////////////////////////////////////////////////////////////
// եåǽ
// GraphConf *cnf : եå
////////////////////////////////////////////////////////////////

void PcgDun::initGraphConf( GraphConf *cnf )
{
//@@@	WSCstring wall_patern;
//	WSCstring door_patern;
//	WSCstring signboard_patern;

	// 롦

	nTileSizeX = cnf->getNum( GRAPH_CONF_TOKEN_TILE_SIZE_X );
	nTileSizeY = cnf->getNum( GRAPH_CONF_TOKEN_TILE_SIZE_Y );

#ifdef D_GTK
	// Ψ
	nTileSizeRate = cnf->getNum( GRAPH_CONF_TOKEN_SCALE );
	g_graph_scale_rate = nTileSizeRate;
#endif // D_GTK

#ifdef D_MFC
	// ŪʳΨ
	nTileSizeRate = _100_PERCENT;
	g_graph_scale_rate = nTileSizeRate;
	// ºݤγΨ
	g_Dir3d.nTileSizeRate
			= cnf->getNum( GRAPH_CONF_TOKEN_SCALE );
#endif // D_MFC

	// 饯Υѥ󡦥

	nPtnDirMbr = N_PTN_DIR_CHR_1;
	nPtnDirMnstr = N_PTN_DIR_CHR_1;

	WSCstring pt;
	pt = cnf->getStr( GRAPH_CONF_TOKEN_CHARACTER_PATERN );
	if( pt == "tile" ){
		nPtnDirMbr = N_PTN_DIR_CHR_1;
		nPtnDirMnstr = N_PTN_DIR_CHR_1;
	}
	if( pt == "RPG_XP" ){
		nPtnDirMbr = N_PTN_DIR_CHR_4;
		nPtnDirMnstr = N_PTN_DIR_CHR_4;
	}
	if( pt == "RPG_2000" ){
		nPtnDirMbr = N_PTN_DIR_CHR_4;
		nPtnDirMnstr = N_PTN_DIR_CHR_4;
	}

	// PC Υѥ󡦥פǾ

	pt = cnf->getStr( GRAPH_CONF_TOKEN_PC_PATERN );
	if( pt == "tile" )
		nPtnDirMbr = N_PTN_DIR_CHR_1;
	if( pt == "RPG_XP" )
		nPtnDirMbr = N_PTN_DIR_CHR_4;
	if( pt == "RPG_2000" )
		nPtnDirMbr = N_PTN_DIR_CHR_4;

	// 󥹥Υѥ󡦥פǾ

	pt = cnf->getStr( GRAPH_CONF_TOKEN_MONSTER_PATERN );
	if( pt == "tile" )
		nPtnDirMnstr = N_PTN_DIR_CHR_1;
	if( pt == "RPG_XP" )
		nPtnDirMnstr = N_PTN_DIR_CHR_4;
	if( pt == "RPG_2000" )
		nPtnDirMnstr = N_PTN_DIR_CHR_4;

	// ƥȡ⡼

	if( cnf->getNum( GRAPH_CONF_TOKEN_TEXT_MODE ) != 0 )
		g_flg_text_mode = TRUE;
	else
		g_flg_text_mode = FALSE;

	// ƥȡ⡼ɻΥեȥ

#ifdef D_GTK
	nTextModeDefaultPoint = cnf->getNum(
			GRAPH_CONF_TOKEN_FONT_SIZE_POINT );
#endif // D_GTK
#ifdef D_MFC
	nTextModeDefaultPoint = cnf->getNum(
			GRAPH_CONF_TOKEN_FONT_SIZE_DOT );
#endif // D_MFC
	g_text_mode_point = nTextModeDefaultPoint;

	nTextSizeX = nTileSizeX;
	nTextSizeY = nTileSizeY;

#ifdef D_MFC
	if( g_flg_text_mode )
		bFlgEnaDrawTurn = false;
	else
		bFlgEnaDrawTurn = true;
#else // D_MFC
	bFlgEnaDrawTurn = true;
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ꥻå
////////////////////////////////////////////////////////////////

void PcgDun::reset()
{
	resetUpdateFlg();

	if( g_flg_text_mode )
		return;

	for( long i = 0; i < PCG_MNSTR_MAX_N; i++ ){
		if( gPcgTabMnstr[i].nRatio <= -1 )
			break;
		if( gPcgTabMnstr[i].nMnstrKind == MNSTR_KIND_NULL )
			break;

		for( long j = 0; j < PCG_LS_MAX_N; j++ ){
			aMnstr[i][j].dis();
		}
	}

	if( get_dun()->lev == 0 ){
		resetMap( gPcgTabNullTown, aaMapNull );
		resetMap( gPcgTabFloorTown, aaMapFloor );
		resetMap( gPcgTabWallTown, aaMapWall );
		resetMap( gPcgTabSignboard, aaMapSignboard );
		resetMap( gPcgTabStairsUpTown, aaMapStairsUp );
		resetMap( gPcgTabStairsDownTown, aaMapStairsDown );
		resetMap( gPcgTabDoorCloseTown, aaMapDoorClose );
		resetMap( gPcgTabDoorOpenTown, aaMapDoorOpen );
	} else {
		resetMap( gPcgTabNullDun, aaMapNull );
		resetMap( gPcgTabFloorDun, aaMapFloor );
		resetMap( gPcgTabWallDun, aaMapWall );
		resetMap( gPcgTabSignboard, aaMapSignboard );
		resetMap( gPcgTabStairsUpDun, aaMapStairsUp );
		resetMap( gPcgTabStairsDownDun, aaMapStairsDown );
		resetMap( gPcgTabDoorCloseDun, aaMapDoorClose );
		resetMap( gPcgTabDoorOpenDun, aaMapDoorOpen );
	}
}

////////////////////////////////////////////////////////////////
// ĤΥեȤꥻå
////////////////////////////////////////////////////////////////

void PcgDun::resetSignboardFont()
{
	nSignboardIntervalWidth = (4 * 2) * nTileSizeRate / _100_PERCENT;
	nSignboardIntervalHeight = (4 * 2) * nTileSizeRate / _100_PERCENT;

#ifdef D_GTK
	// ǥեȤΥ

	if( pSignboardStyle == NULL ){
		pSignboardStyle = gtk_style_copy(
				gtk_widget_get_default_style() );
	}
	if( pSignboardStyle->fg_gc[GTK_STATE_NORMAL] == NULL ){
		pSignboardStyle->fg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}
	if( pSignboardStyle->bg_gc[GTK_STATE_NORMAL] == NULL ){
		pSignboardStyle->bg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}

	// եȤΥݥȿѴ

	nSignboardPoint = 80;
	WSCstring sFontName;

	// 롦˹礦եȡʬõ

	long pntMin = 80;
	long pntMax = 320;
	long pnt = pntMin;
	while( (pntMin + 1) < pntMax ){
		pnt = (pntMin + pntMax) / 2;
		if( setSignboardFont( pnt, &sFontName ) )
			pntMin = pnt;
		else
			pntMax = pnt;
	}
	nSignboardPoint = pntMin;

	setSignboardFont( nSignboardPoint, &sFontName );
#endif // D_GTK

#ifdef D_MFC
	LONG h = getTileSizeY() - nSignboardIntervalHeight;
	h = h * g_Dir3d.nTileSizeRate / _100_PERCENT;
	mSignboardFontDesc.lfHeight = h;
	mSignboardFontDesc.lfWidth = 0;
	mSignboardFontDesc.lfEscapement = 0;
	mSignboardFontDesc.lfOrientation = 0;
	mSignboardFontDesc.lfWeight = FW_NORMAL;
	mSignboardFontDesc.lfItalic = FALSE;
	mSignboardFontDesc.lfUnderline = FALSE;
	mSignboardFontDesc.lfStrikeOut = FALSE;
	mSignboardFontDesc.lfCharSet = SHIFTJIS_CHARSET;
	mSignboardFontDesc.lfOutPrecision
			= OUT_DEFAULT_PRECIS;
	mSignboardFontDesc.lfClipPrecision
			= CLIP_DEFAULT_PRECIS;
	mSignboardFontDesc.lfQuality = DEFAULT_QUALITY;
	mSignboardFontDesc.lfPitchAndFamily
			= (FIXED_PITCH | FF_MODERN);
	strcpy( mSignboardFontDesc.lfFaceName, "Pica" );

	pSignboardFont = new CFont;
	pSignboardFont->CreateFontIndirect(
			&mSignboardFontDesc );

	mD3Signboard.Create( &mSignboardFontDesc );
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ĤΥեȤ
// long pnt : եȤΥݥȿ
// return : եȤॿ륵˼ޤä?
////////////////////////////////////////////////////////////////

bool PcgDun::setSignboardFont( long pnt, WSCstring *pFontName )
{
#ifdef D_GTK
	WSCstring sFontPoint = pnt;
	WSCstring sFontName = "";
	sFontName += "-*-*-*-*-normal--*-";
	sFontName += sFontPoint;
	sFontName += "-*,-*";

	*pFontName = sFontName;

	gtk_style_set_font( pSignboardStyle, gdk_fontset_load( sFontName ) );

	// եȡ

	GdkFont *font = gtk_style_get_font( pSignboardStyle );
	if( font == NULL )
		return false;
	gdk_string_extents( font, "M",
			&nSignboardLeftBearing, &nSignboardRightBearing,
			&nSignboardWidth,
			&nSignboardAscent, &nSignboardDescent );

	nSignboardSizeX = nSignboardLeftBearing + nSignboardRightBearing
			+ nSignboardIntervalWidth;
	nSignboardSizeY = nSignboardAscent + nSignboardDescent
			+ nSignboardIntervalHeight;

	// եȡå

	if( nSignboardSizeX > getTileSizeX( false ) )
		return false;
	if( nSignboardSizeY > getTileSizeY( false ) )
		return false;
#endif // D_GTK

	return true;
}

////////////////////////////////////////////////////////////////
// ѥοʿδॵ֤
// return : ॵ
////////////////////////////////////////////////////////////////

long PcgDun::getTileSizeX( bool flgText )
{
	long n = 1;

	if( g_flg_text_mode && flgText )
		n = nTextSizeX * nTileSizeRate / _100_PERCENT;
	else
		n = nTileSizeX * nTileSizeRate / _100_PERCENT;

	if( n < 1 )
		n = 1;

	return n;
}

////////////////////////////////////////////////////////////////
// ѥοľδॵ֤
// return : ॵ
////////////////////////////////////////////////////////////////

long PcgDun::getTileSizeY( bool flgText )
{
	long n = 1;

	if( g_flg_text_mode && flgText )
		n = nTextSizeY * nTileSizeRate / _100_PERCENT;
	else
		n = nTileSizeY * nTileSizeRate / _100_PERCENT;

	if( n < 1 )
		n = 1;

	return n;
}

////////////////////////////////////////////////////////////////
// ƥȡ⡼ɤΥեȡ֤
// return : եȡ
////////////////////////////////////////////////////////////////

long PcgDun::getTextModePoint()
{
	long n = g_text_mode_point * nTileSizeRate / _100_PERCENT;

	if( n < 1 )
		n = 1;

	return n;
}

////////////////////////////////////////////////////////////////
// ѥΥХꥨ󡦥ޥåפΥꥻå
// PcgTab tab[] : ѥΥơ֥
// PcgMap map[MAP_MAX_Y][MAP_MAX_X] : ХꥨΥޥå
////////////////////////////////////////////////////////////////

void PcgDun::resetMap( PcgTab tab[], PcgMap map[MAP_MAX_Y][MAP_MAX_X] )
{
	long nRatioSum = 0;
	for( long i = 0; i < LOOP_MAX_100; i++ ){
		if( tab[i].nRatio <= -1 )
			break;

		nRatioSum += tab[i].nRatio;
	}

	for( long y = 0; y < MAP_MAX_Y; y++ ){
		for( long x = 0; x < MAP_MAX_X; x++ ){
			map[y][x] = 0;

			long n = randm( nRatioSum );
			long nSum = 0;
			long i = 0;
			for( i = 0; i < LOOP_MAX_100; i++ ){
				if( tab[i].nRatio <= -1 )
					break;

				nSum += tab[i].nRatio;
				if( n < nSum ){
					map[y][x] = (PcgMap)i;
					break;
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////
// ֥롦Хåե֤
// return : ե졼ࡦХåե
////////////////////////////////////////////////////////////////

WSDmwindowDev *PcgDun::getWBuf()
{
	if( pWBuf == NULL ){
		pWBuf = WSDmwindowDev::getNewInstance();
		pWBuf->createPixmap(
				MAP_MAX_X * getTileSizeX( true ),
				MAP_MAX_Y * getTileSizeY( true ) );
	}

	return pWBuf;
}

////////////////////////////////////////////////////////////////
// ֥롦Хåեݤʤ
// return : ե졼ࡦХåե
////////////////////////////////////////////////////////////////

WSDmwindowDev *PcgDun::ReNewWBuf()
{
#ifdef D_GTK
	if( pWBuf == NULL ){
		getWBuf();
		return pWBuf;
	}

	gdk_pixmap_unref( pWBuf->getPixMap() );

	pWBuf->createPixmap(
			MAP_MAX_X * getTileSizeX( true ),
			MAP_MAX_Y * getTileSizeY( true ) );
#endif // D_GTK

	return pWBuf;
}

////////////////////////////////////////////////////////////////
// ޥåפΥʥκɸǻꤹ
// long mapX : X ɸ
// long mapY : Y ɸ
////////////////////////////////////////////////////////////////

void PcgDun::scrollTile( long mapX, long mapY )
{
	scrollMap(
			mapX * getTileSizeX( true ),
			mapY * getTileSizeY( true ) );
}

////////////////////////////////////////////////////////////////
// ޥåפΥʥɥåȤκɸǻꤹ
// long x : X ɸ
// long y : Y ɸ
////////////////////////////////////////////////////////////////

void PcgDun::scrollMap( long x, long y )
{
	long maxX = getScrollBarMaxX();
	long maxY = getScrollBarMaxY();
	long w = getScrollBarW();
	long h = getScrollBarH();

	if( x < 0 )
		x = 0;
	if( y < 0 )
		y = 0;
	if( x > maxX )
		x = maxX;
	if( y > maxY )
		y = maxY;

	setScrollBarX( x );
	setScrollBarY( y );

	drawScroll( x, y, w, h );
}

////////////////////////////////////////////////////////////////
// ޥåפ뤷褹ɬפͭ뤫å
// long x : X ɸ
// long y : Y ɸ
// long w : 
// long h : ⤵
// return : 褹ɬפͭ뤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkScroll( long x, long y, long w, long h )
{
	x /= getTileSizeX( true );
	y /= getTileSizeY( true );

	static long preX = -1;
	static long preY = -1;
	static long preW = -1;
	static long preH = -1;

	if( (x == preX) && (y == preY) && (w == preW) && (h == preH) )
		return false;

	preX = x;
	preY = y;
	preW = w;
	preH = h;

	return true;
}

////////////////////////////////////////////////////////////////
// ޥåפ뤷
// long x : X ɸ
// long y : Y ɸ
// long w : 
// long h : ⤵
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawScroll( long x, long y, long w, long h )
{
	if( chk_nest_flg_dun() )
		return false;

	sort_all_chr();

	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );

	long mapX1 = x / sizX;
	long mapY1 = y / sizY;
	long mapX2 = (x + w + sizX - 1) / sizX;
	long mapY2 = (y + h + sizY - 1) / sizY;
	long mapW = mapX2 - mapX1;
	long mapH = mapY2 - mapY1;

	for( long yy = mapY1; yy < mapY2; yy++ ){
		for( long xx = mapX1; xx < mapX2; xx++ ){
			if( !getUpdateFlg( xx, yy ) )
				continue;

			if( !drawTile( xx, yy, false ) )
				return false;
		}
	}

#ifdef D_MFC
	if( g_Dir3d.GetFlagRefresh() ){
		flush( mapX1, mapY1, mapW, mapH );
	}

	setUpdateFlgRange( mapX1, mapY1, mapW, mapH, true );
	g_Dir3d.SetFlagRefresh( +1 );
#else // D_MFC
	flush( mapX1, mapY1, mapW, mapH );
#endif // D_MFC

	return true;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦Сξ֤
// SCROLLINFO *info : ֤
// int sb : 롦Сμ
////////////////////////////////////////////////////////////////

#ifdef D_MFC
void PcgDun::getScrollInfo( SCROLLINFO *info, int sb )
{
	if( info == NULL )
		return;
	if( theApp.m_pMainFrm == NULL )
		return;

	HWND hWnd = theApp.m_pMainFrm->GetViewWnd();

	info->cbSize = sizeof( *info );
	info->fMask = SIF_ALL;
	::GetScrollInfo( hWnd, sb, info );
}
#endif // D_MFC

////////////////////////////////////////////////////////////////
// ޥåפΥ롦Сΰ֤
// int sb : 롦Сμ
// int pos : 롦Сΰ
////////////////////////////////////////////////////////////////

#ifdef D_MFC
void PcgDun::setScrollInfoPos( int sb, int pos )
{
	if( theApp.m_pMainFrm == NULL )
		return;

	HWND hWnd = theApp.m_pMainFrm->GetViewWnd();

	SCROLLINFO info;
	info.cbSize = sizeof( info );
	info.fMask = SIF_POS;
	info.nPos = pos;

	SetScrollInfo( hWnd, sb, &info, TRUE );
}
#endif // D_MFC

////////////////////////////////////////////////////////////////
// ޥåפΥ롦С X ɸ֤
// return : X ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarX()
{
	long x = 0;

#ifdef D_WS
	x = (long)MapScrollX->getProperty( WSNvalue );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *hAdj = GTK_ADJUSTMENT( gMapHScrollBarAdjustment );

	x = (long)hAdj->value;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_HORZ );

	x = info.nPos;
#endif // D_MFC

	return x;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦С Y ɸ֤
// return : Y ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarY()
{
	long y = 0;

#ifdef D_WS
	y = (long)MapScrollY->getProperty( WSNvalue );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *vAdj = GTK_ADJUSTMENT( gMapVScrollBarAdjustment );

	y = (long)vAdj->value;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_VERT );

	y = info.nPos;
#endif // D_MFC

	return y;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦С֤
// return : 
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarW()
{
	long w = 1;

#ifdef D_WS
	w = (long)MapScrollX->getProperty( WSNsliderSize );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *hAdj = GTK_ADJUSTMENT( gMapHScrollBarAdjustment );

	w = (long)hAdj->page_size;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_HORZ );

	w = info.nPage;
	if( g_Dir3d.nTileSizeRate > 0 )
		w = w * _100_PERCENT / g_Dir3d.nTileSizeRate;
#endif // D_MFC

	return w;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦Сι⤵֤
// return : ⤵
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarH()
{
	long h = 1;

#ifdef D_WS
	h = (long)MapScrollY->getProperty( WSNsliderSize );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *vAdj = GTK_ADJUSTMENT( gMapVScrollBarAdjustment );

	h = (long)vAdj->page_size;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_VERT );

	h = info.nPage;
	if( g_Dir3d.nTileSizeRate > 0 )
		h = h * _100_PERCENT / g_Dir3d.nTileSizeRate;
#endif // D_MFC

	return h;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦Сκ X ɸ֤
// return :  X ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarMaxX()
{
	long maxX = 1;

#ifdef D_WS
	maxX = (long)MapScrollX->getProperty( WSNmaximum );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *hAdj = GTK_ADJUSTMENT( gMapHScrollBarAdjustment );

	long w = (long)hAdj->page_size;
	maxX = (long)hAdj->upper - w;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_HORZ );

	maxX = info.nMax - info.nPage + 2;
#endif // D_MFC

	return maxX;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦Сκ Y ɸ֤
// return :  Y ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getScrollBarMaxY()
{
	long maxY = 1;

#ifdef D_WS
	maxY = (long)MapScrollY->getProperty( WSNmaximum );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *vAdj = GTK_ADJUSTMENT( gMapVScrollBarAdjustment );

	long h = (long)vAdj->page_size;
	maxY = (long)vAdj->upper - h;
#endif // D_GTK

#ifdef D_MFC
	SCROLLINFO info;
	getScrollInfo( &info, SB_VERT );

	maxY = info.nMax - info.nPage + 1;
#endif // D_MFC

	return maxY;
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦С X ɸ
////////////////////////////////////////////////////////////////

void PcgDun::setScrollBarX( long x )
{
#ifdef D_WS
	MapScrollX->setProperty( WSNvalue, x );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *hAdj = GTK_ADJUSTMENT( gMapHScrollBarAdjustment );
	hAdj->value = x;
	gtk_adjustment_value_changed( hAdj );
#endif // D_GTK

#ifdef D_MFC
	setScrollInfoPos( SB_HORZ, x );
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ޥåפΥ롦С Y ɸ
////////////////////////////////////////////////////////////////

void PcgDun::setScrollBarY( long y )
{
#ifdef D_WS
	MapScrollY->setProperty( WSNvalue, y );
#endif // D_WS

#ifdef D_GTK
	GtkAdjustment *vAdj = GTK_ADJUSTMENT( gMapVScrollBarAdjustment );
	vAdj->value = y;
	gtk_adjustment_value_changed( vAdj );
#endif // D_GTK

#ifdef D_MFC
	setScrollInfoPos( SB_VERT, y );
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// κɸޥåפΥɥϰ⤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : ϰ⤫
////////////////////////////////////////////////////////////////

bool PcgDun::clipWin( long mapX, long mapY )
{
	return( clipWinX( mapX ) && clipWinY( mapY ) );
}

////////////////////////////////////////////////////////////////
//  X ɸޥåפΥɥϰ⤫Ĵ٤
// long mapX : X ɸ
// return : ϰ⤫
////////////////////////////////////////////////////////////////

bool PcgDun::clipWinX( long mapX )
{
	long minX = getScrollBarX();
	long maxX = (getScrollBarW() - 1) + minX;

	long x1 = mapX * getTileSizeX( true );
	long x2 = x1 + (getTileSizeX( true ) - 1);

	if( x2 < minX )
		return false;
	if( x1 > maxX )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
//  Y ɸޥåפΥɥϰ⤫Ĵ٤
// long mapX : Y ɸ
// return : ϰ⤫
////////////////////////////////////////////////////////////////

bool PcgDun::clipWinY( long mapY )
{
	long minY = getScrollBarY();
	long maxY = (getScrollBarH() - 1) + minY;

	long y1 = mapY * getTileSizeY( true );
	long y2 = y1 + (getTileSizeY( true ) - 1);

	if( y2 < minY )
		return false;
	if( y1 > maxY )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// ޥåפκ
// bool flgForce : 
////////////////////////////////////////////////////////////////

void PcgDun::redraw( bool flgForce )
{
	if( bFlgEnaDrawTurn && !flgForce )
		return;

	setUpdateFlgRange( 0, 0, MAP_MAX_X, MAP_MAX_Y, true );
	drawTurn( true );
}

////////////////////////////////////////////////////////////////
// ޥåפ
// long mapX : X ɸ
// long mapY : Y ɸ
// long mapW : 
// long mapH : ⤵
// bool flgForce : 
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::draw(
	long mapX, long mapY,
	long mapW, long mapH,
	bool flgForce
)
{
	setUpdateFlgRange( mapX, mapY, mapW, mapH, true );

	if( bFlgEnaDrawTurn && !flgForce )
		return true;
	if( chk_nest_flg_dun() )
		return false;

#ifdef D_MFC
	if( g_Dir3d.GetFlagRefresh() ){
		drawSub( mapX, mapY, mapW, mapH );
		flush( mapX, mapY, mapW, mapH );
	}
#else // D_MFC
	drawSub( mapX, mapY, mapW, mapH );
	flush( mapX, mapY, mapW, mapH );
#endif // D_MFC

#ifdef D_MFC
	g_Dir3d.SetFlagRefresh( +1 );
#endif // D_MFC

	return true;
}

////////////////////////////////////////////////////////////////
// Υޥåפ
// bool flgForce : 
////////////////////////////////////////////////////////////////

void PcgDun::drawTurn( bool flgForce )
{
	long x = getScrollBarX();
	long y = getScrollBarY();
	long w = getScrollBarW();
	long h = getScrollBarH();

	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );

	long mapX1 = x / sizX;
	long mapY1 = y / sizY;
	long mapX2 = (x + w + (sizX - 1)) / sizX;
	long mapY2 = (y + h + (sizY - 1)) / sizY;
	long mapW = mapX2 - mapX1;
	long mapH = mapY2 - mapY1;

	if( !bFlgEnaDrawTurn && !flgForce )
		return;

	if( chk_nest_flg_dun() ){
		resetUpdateFlg();

		return;
	}

#ifdef D_MFC
	if( g_Dir3d.GetFlagRefresh() ){
		drawSub( mapX1, mapY1, mapW, mapH );
		flush( mapX1, mapY1, mapW, mapH );
	}
#else // D_MFC
	drawSub( mapX1, mapY1, mapW, mapH );
	flush( mapX1, mapY1, mapW, mapH );
#endif // D_MFC

#ifdef D_MFC
	g_Dir3d.SetFlagRefresh( +1 );
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ޥåפζ̴ؿ
// long mapX : X ɸ
// long mapY : Y ɸ
// long mapW : 
// long mapH : ⤵
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawSub( long mapX, long mapY, long mapW, long mapH )
{
	if( g_flg_text_mode )
		return drawText( mapX, mapY, mapW, mapH );

	long mapX1 = mapX - MAP_DRAW_ADD_R;
	long mapY1 = mapY - MAP_DRAW_ADD_R;
	long mapX2 = mapX + mapW + MAP_DRAW_ADD_R;
	long mapY2 = mapY + mapH + MAP_DRAW_ADD_R;

	long objX1 = mapX;
	long objY1 = mapY;
	long objX2 = mapX + mapW;
	long objY2 = mapY + mapH;

	long sx = mapX1 * getTileSizeX();
	long sy = mapY1 * getTileSizeY();
	long sw = (mapX2 - mapX1) * getTileSizeX();
	long sh = (mapY2 - mapY1) * getTileSizeY();

	// ƥե饰

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) ){
				setUpdateFlg( x, y, true );
			}
		}
	}

	// ֥Ȥ

	for( long y = objY1; y < objY2; y++ ){
		for( long x = objX1; x < objX2; x++ ){
			if( !getUpdateFlg( x, y ) )
				continue;
			if( !clipWin( x, y ) )
				continue;

#ifdef D_MFC
			drawObj( x, y );
			setUpdateFlg( x, y, true );
#else // D_MFC
			if( drawObj( x, y ) )
				setUpdateFlg( x, y, false );
			else
				setUpdateFlg( x, y, true );
#endif // D_MFC
		}
	}

	// 櫤

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) )
				continue;

			if( !drawTrap( x, y ) )
				setUpdateFlg( x, y, true );
		}
	}

	// ʤ

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) )
				continue;

			if( !drawStairsUp( x, y ) )
				setUpdateFlg( x, y, true );
			if( !drawStairsDown( x, y ) )
				setUpdateFlg( x, y, true );
		}
	}

	// 󥫡

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) )
				continue;

			if( !drawSquare( x, y ) )
				setUpdateFlg( x, y, true );
		}
	}

	// ƥ

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) )
				continue;

			if( !drawItemAll( x, y ) )
				setUpdateFlg( x, y, true );
		}
	}

	// 饯

	drawChrListAll( sx, sy, sw, sh );

	// 

	for( long y = mapY1; y < mapY2; y++ ){
		for( long x = mapX1; x < mapX2; x++ ){
			if( !clipWin( x, y ) )
				continue;

			if( !drawCrsrAll( x, y ) )
				setUpdateFlg( x, y, true );
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////
// ե졼ࡦХåե饦ɥؤ
// long mapX : X ɸ
// long mapY : Y ɸ
// long mapW : 
// long mapH : ⤵
////////////////////////////////////////////////////////////////

void PcgDun::flush( long mapX, long mapY, long mapW, long mapH )
{
	if( chk_nest_flg_dun() )
		return;

	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );

	long x1 = mapX * sizX;
	long y1 = mapY * sizY;
	long w = mapW * sizX;
	long h = mapH * sizY;
	long x2 = x1 + w;
	long y2 = y1 + h;

	long minX = getScrollBarX();
	long minY = getScrollBarY();
	long maxX = minX + getScrollBarW();
	long maxY = minY + getScrollBarH();

	if( x1 < minX )
		x1 = minX;
	if( y1 < minY )
		y1 = minY;
	if( x1 > maxX )
		x1 = maxX;
	if( y1 > maxY )
		y1 = maxY;

	if( x2 < minX )
		x2 = minX;
	if( y2 < minY )
		y2 = minY;
	if( x2 > maxX )
		x2 = maxX;
	if( y2 > maxY )
		y2 = maxY;

	w = x2 - x1;
	h = y2 - y1;
	if( w <= 0 )
		return;
	if( h <= 0 )
		return;

	if( !getWBuf()->getReady() )
		return;
#ifdef D_WS
	if( !MapForm->getdev()->getReady() )
		return;

	getWBuf()->copyToWindow( MapForm->getdev(),
			x1, y1, w, h,
			x1 - minX, y1 - minY );
#endif // D_WS

#ifdef D_GTK
	GdkRectangle updateRect;
	updateRect.x = x1 - minX;
	updateRect.y = y1 - minY;
	updateRect.width = w;
	updateRect.height = h;
	gtk_widget_draw( gMapDrawingArea, &updateRect );
#endif // D_GTK

#ifdef D_MFC
#endif // D_MFC
}

////////////////////////////////////////////////////////////////
// ƥȡ⡼ɤν
////////////////////////////////////////////////////////////////

void PcgDun::initText()
{
#ifdef D_GTK
	if( g_text_mode_point <= 0 )
		g_text_mode_point = nTextModeDefaultPoint;

	// ǥեȤΥ

	if( pStyle == NULL )
		pStyle = gtk_style_copy( gtk_widget_get_default_style() );
	if( pStyle->fg_gc[GTK_STATE_NORMAL] == NULL ){
		pStyle->fg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}
	if( pStyle->bg_gc[GTK_STATE_NORMAL] == NULL ){
		pStyle->bg_gc[GTK_STATE_NORMAL]
				= gdk_gc_new( gMapDrawingArea->window );
	}

	// եȤ

	WSCstring sFontPoint = getTextModePoint() * 10;
	WSCstring sFontName = "";
	sFontName += "-*-*-*-r-normal--*-";
	sFontName += sFontPoint;
	sFontName += "-*,-*";

	gtk_style_set_font( pStyle, gdk_fontset_load( sFontName ) );

#if 0
//@@@
	sFontName = user_select_font();
	if( sFontName != "" ){
		gtk_style_set_font( pStyle,
				gdk_fontset_load( sFontName ) );
	}
#elif 0
	GdkFont *selFont = NULL;
	user_select_font( &selFont );
	gtk_style_set_font( pStyle, selFont );

#elif 0
//@@@
	GdkFont *font = NULL;
	user_select_font( &font );
	gtk_style_set_font( pStyle, font );
#endif

	// եȡ

	GdkFont *font = gtk_style_get_font( pStyle );
	gdk_string_extents( font, "M",
			&nLeftBearing, &nRightBearing,
			&nWidth, &nAscent, &nDescent );

#if FLAG_FONT_TO_TEXT_SIZE
	nTextSizeX = nLeftBearing + nRightBearing + nIntervalWidth;
	nTextSizeX *= 2;
	nTextSizeY = nAscent + nDescent + nIntervalHeight;

	nTextSizeX = nTextSizeX * _100_PERCENT / nTileSizeRate;
	nTextSizeY = nTextSizeY * _100_PERCENT / nTileSizeRate;
#endif

	// 롦СĴ

	GtkAdjustment *adj;
	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );
	long w = gMapDrawingArea->allocation.width;
	long h = gMapDrawingArea->allocation.height;

	adj = GTK_ADJUSTMENT( gMapHScrollBarAdjustment );
	adj->lower = 0;
	adj->upper = MAP_MAX_X * sizX;
	adj->step_increment = 1 * sizX;
	adj->page_increment = AREA_MAX_X / 2 * sizX;
	adj->page_size = w;
	gtk_adjustment_changed( adj );

	adj = GTK_ADJUSTMENT( gMapVScrollBarAdjustment );
	adj->lower = 0;
	adj->upper = MAP_MAX_Y * sizY;
	adj->step_increment = 1 * sizY;
	adj->page_increment = AREA_MAX_Y / 2 * sizY;
	adj->page_size = h;
	gtk_adjustment_changed( adj );
#endif // D_GTK

#ifdef D_MFC
	// ǥեȤΥ

	if( pStyle == NULL )
		pStyle = (void *)1;

	if( g_text_mode_point <= 0 )
		g_text_mode_point = nTextModeDefaultPoint;

	CXlnlView *view = theApp.m_pMainFrm->GetView();

	// եȡ

#if FLAG_FONT_TO_TEXT_SIZE
	CClientDC dc( view );

	CSize siz = dc.GetTextExtent( "M", 1 );

	nTextSizeX = siz.cx;
	nTextSizeX *= 2;
	nTextSizeY = siz.cy;

	nTextSizeX = nTextSizeX * _100_PERCENT / nTileSizeRate;
	nTextSizeY = nTextSizeY * _100_PERCENT / nTileSizeRate;
#endif

	// եȤ

	LONG h = g_text_mode_point;
	h = h * g_Dir3d.nTileSizeRate / _100_PERCENT;
	mTextFontDesc.lfHeight = h;
	mTextFontDesc.lfWidth = 0;
	mTextFontDesc.lfEscapement = 0;
	mTextFontDesc.lfOrientation = 0;
	mTextFontDesc.lfWeight = FW_NORMAL;
	mTextFontDesc.lfItalic = FALSE;
	mTextFontDesc.lfUnderline = FALSE;
	mTextFontDesc.lfStrikeOut = FALSE;
	mTextFontDesc.lfCharSet = SHIFTJIS_CHARSET;
	mTextFontDesc.lfOutPrecision
			= OUT_DEFAULT_PRECIS;
	mTextFontDesc.lfClipPrecision
			= CLIP_DEFAULT_PRECIS;
	mTextFontDesc.lfQuality = DEFAULT_QUALITY;
	mTextFontDesc.lfPitchAndFamily
			= (FIXED_PITCH | FF_MODERN);
	strcpy( mTextFontDesc.lfFaceName, "Pica" );

	pTextFont = new CFont;
	pTextFont->CreateFontIndirect(
			&mTextFontDesc );

	// 롦СĴ

	view->ResetScroll();
#endif // MFC
}

////////////////////////////////////////////////////////////////
// ƥȡ⡼ɤΥޥåפ
// long mapX : X ɸ
// long mapY : Y ɸ
// long mapW : 
// long mapH : ⤵
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawText( long mapX, long mapY, long mapW, long mapH )
{
#ifdef D_GTK
	if( pStyle == NULL )
		initText();

	dun_t *dun = get_dun();
	if( dun == NULL )
		return false;

	pos_t *pos = get_map_draw_pos();
	if( pos == NULL )
		return false;

	if( gMapDrawingArea == NULL )
		return false;

	if( chk_nest_flg_dun() )
		return true;

	GdkDrawable *d = getWBuf()->getPixMap();
	GdkFont *font = gtk_style_get_font( pStyle );
	GdkGC *fgGc = pStyle->fg_gc[GTK_STATE_NORMAL];
	GdkGC *bgGc = pStyle->bg_gc[GTK_STATE_NORMAL];

	set_map_total( mapX, mapY, mapW, mapH );
	char s[4 + 1];
	s[0] = ' ';
	s[1] = '\0';

	for( long h = 0; h < mapH; h++ ){
		for( long w = 0; w < mapW; w++ ){
			long mx = mapX + w;
			long my = mapY + h;

			if( !clip_pos( mx, my ) )
				continue;

			setUpdateFlg( mx, my, true );

			if( !clipWin( mx, my ) )
				continue;

			// ɸ׻

			long mjrX = mx * getTileSizeX( true );
			long mnrX = mjrX + (getTileSizeX( true ) / 2);
			long y = my * getTileSizeY( true );

			// °

			curs_attr_t attr = dun->map.attr[my][mx];
			curs_attrset_dir( &attr );
			SetPenAttr( fgGc, bgGc, &attr );

			// طʤɤĤ֤

			gdk_draw_rectangle( d, bgGc, TRUE,
					mjrX, y,
					getTileSizeX( true ),
					getTileSizeY( true ) );

			// 饤

			long yy = y + getTileSizeY( true ) - 1;
			if( attr.attr & A_UNDERLINE ){
				gdk_draw_line( d, fgGc,
						mjrX,
						yy,
						mjrX + getTileSizeX( true ),
						yy );
			}

			// ɸ

			mjrX += nLeftBearing;
			mnrX += nLeftBearing;
			y += nAscent + 1;

			// ʸ

			s[0] = dun->map.total.mjr[my][mx];
			gdk_draw_string( d, font, fgGc, mjrX, y, s );

			s[0] = dun->map.total.mnr[my][mx];
			gdk_draw_string( d, font, fgGc, mnrX, y, s );

			setUpdateFlg( mx, my, false );
		}
	}

	curs_attrset_dflt();
#endif // D_GTK

#ifdef D_MFC
	if( !gFlagInitMfcDone )
		return true;
	if( pStyle == NULL )
		initText();

	dun_t *dun = get_dun();
	if( dun == NULL )
		return false;

	pos_t *pos = get_map_draw_pos();
	if( pos == NULL )
		return false;

	if( chk_nest_flg_dun() )
		return true;

	if( pDc == NULL ){
		CXlnlView *view = theApp.m_pMainFrm->GetView();
		pDc = new CClientDC( view );
	}

	long scrollX = getScrollBarX();
	long scrollY = getScrollBarY();
	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );
	long offsetX = scrollX / sizX;
	long offsetY = scrollY / sizY;

	set_map_total( mapX, mapY, mapW, mapH );
	char s[4 + 1];
	s[0] = ' ';
	s[1] = '\0';

	for( long h = 0; h < mapH; h++ ){
		for( long w = 0; w < mapW; w++ ){
			long mx = mapX + w;
			long my = mapY + h;

			if( !clip_pos( mx, my ) )
				continue;

			setUpdateFlg( mx, my, true );

			if( !clipWin( mx, my ) )
				continue;

			// ɸ׻

			long mjrX = mx - offsetX;
			mjrX *= getTileSizeX( true );

			long mnrX = mjrX + (getTileSizeX( true ) / 2);

			long y = my - offsetY;
			y *= getTileSizeY( true );

			mjrX *= g_Dir3d.nTileSizeRate;
			mjrX /= _100_PERCENT;
			mnrX *= g_Dir3d.nTileSizeRate;
			mnrX /= _100_PERCENT;
			y *= g_Dir3d.nTileSizeRate;
			y /= _100_PERCENT;

			long tw = getTileSizeX( true );
			tw *= g_Dir3d.nTileSizeRate;
			tw /= _100_PERCENT;

			long th = getTileSizeY( true );
			th *= g_Dir3d.nTileSizeRate;
			th /= _100_PERCENT;

			// 賫

			g_Dir3d.BeginDraw();

			// °

			curs_attr_t attr = dun->map.attr[my][mx];
			curs_attrset_dir( &attr );
			COLORREF colorBg;
			CvColorRGB( &colorBg, attr.bg );

			// طʤɤĤ֤

			CRect rect( mjrX, y, mjrX + tw, y + th );
			pDc->FillSolidRect( rect, colorBg );

			// 饤

			long yy = y + th - 1;
			if( attr.attr & A_UNDERLINE ){
				CPen *pen = new CPen();
				SetPenAttrFg( pDc, pen, NULL, &attr );

				pDc->MoveTo( mjrX, yy );
				pDc->LineTo( mjrX + tw, yy );

				delete pen;
			}

			// ɸ

			// ʸ

			CPen *pen = new CPen();
			SetPenAttrFg( pDc, pen, NULL, &attr );
			CFont *pPrevFont = pDc->SelectObject(
					pTextFont );

			s[0] = dun->map.total.mjr[my][mx];
			pDc->TextOut( mjrX, y, s );

			s[0] = dun->map.total.mnr[my][mx];
			pDc->TextOut( mnrX, y, s );

			pDc->SelectObject( pPrevFont );
			delete pen;

			// 轪λ

			g_Dir3d.EndDraw();

			setUpdateFlg( mx, my, false );
		}
	}

	curs_attrset_dflt();
#endif // D_MFC

	return true;
}

////////////////////////////////////////////////////////////////
// եȡ쥯󡦥μ¹
// GdkFont **font : Ф줿եȤ֤
// return : 顼̵ä
////////////////////////////////////////////////////////////////

#ifdef D_GTK
bool PcgDun::user_select_font( GdkFont **font )
{
	if( font == NULL )
		return false;

	GtkWidget *dlg = gtk_font_selection_dialog_new( "Select font" );
	GtkFontSelectionDialog *fontSelDlg
			= GTK_FONT_SELECTION_DIALOG( dlg );

	gtk_signal_connect(
			GTK_OBJECT( fontSelDlg ),
			"destroy",
			GTK_SIGNAL_FUNC( handle_sel_font_destroy ),
			GTK_OBJECT( fontSelDlg ) );
	gtk_signal_connect_object(
			GTK_OBJECT( fontSelDlg->cancel_button ),
			"clicked",
			GTK_SIGNAL_FUNC( gtk_widget_destroy ),
			GTK_OBJECT( fontSelDlg ) );
	gtk_signal_connect(
			GTK_OBJECT( fontSelDlg->ok_button ),
			"clicked",
			GTK_SIGNAL_FUNC( handle_sel_font_ok ),
			GTK_OBJECT( fontSelDlg ) );

	gtk_grab_add( GTK_WIDGET( dlg ) );
	gtk_widget_show( GTK_WIDGET( dlg ) );

	gtk_main();

	if( *font != NULL )
		gdk_font_unref( *font );
	*font = gSelFont;

	return true;
}
#endif // D_GTK

////////////////////////////////////////////////////////////////
// GTK+ νλΥϥɥ
////////////////////////////////////////////////////////////////

#ifdef D_GTK
static void handle_sel_font_destroy(
	GtkWidget *widget, gpointer user_data
)
{
	gtk_main_quit();
	gtk_grab_remove( GTK_WIDGET( user_data ) );
}
#endif // D_GTK

////////////////////////////////////////////////////////////////
// եȡ쥯󡦥
// OK ܥΥå٥Ƚ
// GtkWidget *widget : ܥ
// gpointer data : ϿϤ줿ǡ
////////////////////////////////////////////////////////////////

#ifdef D_GTK
static void handle_sel_font_ok(
	GtkWidget *widget, gpointer *data
)
{
	gSelFont = gtk_font_selection_dialog_get_font(
			GTK_FONT_SELECTION_DIALOG( data ) );

	gdk_font_ref( gSelFont );

	gtk_widget_destroy( GTK_WIDGET( data ) );
}
#endif // D_GTK

////////////////////////////////////////////////////////////////
// ե饰Υꥻå
////////////////////////////////////////////////////////////////

void PcgDun::resetUpdateFlg()
{
	setUpdateFlgRange( 0, 0, MAP_MAX_X, MAP_MAX_Y, true );
}

////////////////////////////////////////////////////////////////
// ϰϤκե饰
// long mapX : X ɸ
// long mapY : Y ɸ
// long mapW : 
// long mapH : ⤵
// bool flg : 褹ɬפ뤫
////////////////////////////////////////////////////////////////

void PcgDun::setUpdateFlgRange(
	long mapX, long mapY,
	long mapW, long mapH,
	bool flg
)
{
	for( long dy = 0; dy < mapH; dy++ )
		for( long dx = 0; dx < mapW; dx++ )
			setUpdateFlg( mapX + dx, mapY + dy, flg );
}

////////////////////////////////////////////////////////////////
// ե饰
// long mapX : X ɸ
// long mapY : Y ɸ
// bool flg : 褹ɬפ뤫
////////////////////////////////////////////////////////////////

void PcgDun::setUpdateFlg( long mapX, long mapY, bool flg )
{
	if( !clip_pos( mapX, mapY ) )
		return;

	bFlgUpdate[mapY][mapX] = flg;
}

////////////////////////////////////////////////////////////////
// ե饰μ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 褹ɬפ뤫
////////////////////////////////////////////////////////////////

bool PcgDun::getUpdateFlg( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return false;

#ifdef D_MFC
	if( g_flg_text_mode )
		return true;
#endif // D_MFC

	return bFlgUpdate[mapY][mapX];
}

////////////////////////////////////////////////////////////////
// ȥ̤
////////////////////////////////////////////////////////////////

bool PcgDun::drawTitle()
{
	if( nTitleRandmSelN <= -1 )
		return false;

	// طʤɤĤ֤

#ifdef D_GTK
	GdkDrawable *d = getWBuf()->getPixMap();
	GdkGC *gc = pTitleStyle->bg_gc[GTK_STATE_NORMAL];
	gdk_draw_rectangle( d, gc, TRUE,
			0, 0,
			gMapDrawingArea->allocation.width,
			gMapDrawingArea->allocation.height );
#endif // D_GTK

#ifdef D_MFC
#endif // D_MFC

	// طʲ

	long x = gPcgDun.getScrollBarX();
	long y = gPcgDun.getScrollBarY();
	long w = pcgTitle.getWidth();
	long h = pcgTitle.getHeight();
	pcgTitle.drawFix( getWBuf(), x, y, w, h );

	// ȥ롦͡

	static time_t t = 0;
	static long frame = 0;
	static drawTitleKindType kind = DRAW_TITLE_KIND_1;
	if( time( NULL ) >= t ){
		frame = 0;
		kind = (drawTitleKindType)randm( DRAW_TITLE_KIND_MAX_N );
	}
	if( frame <= 0 )
		t = time( NULL ) + 10;

	switch( kind ){
	case DRAW_TITLE_KIND_0:
		drawTitle0( frame );
		break;
	case DRAW_TITLE_KIND_1:
		drawTitle1( frame );
		break;
	case DRAW_TITLE_KIND_MAX_N:
		break;
	}
#ifdef D_MFC
	g_Dir3d.SetFlagRefresh( +1 );
#endif // D_MFC

	frame++;

	return true;
}

void PcgDun::drawTitle0( long frame )
{
	if( pTitleStyle == NULL )
		initTitle();

	long len = 1;
	for( long j = 0; j < 3; j++ )
		len = max_l( len, str_len_draw( gStrTitle[j] ) );

#ifdef D_GTK
	const long bgnX = gMapDrawingArea->allocation.width;
	const long baseX = (gMapDrawingArea->allocation.width
			- (len * TITLE_FONT_DOT)) / 2;
	const long baseY = gMapDrawingArea->allocation.height / 2;
#endif // D_GTK

#ifdef D_MFC
	CXlnlView *view = theApp.m_pMainFrm->GetView();
	CRect rect;
	view->GetClientRect( &rect );

	const long bgnX = rect.Width();
	const long baseX = (rect.Width() / 2)
			- (len * TITLE_FONT_DOT / 2);
	const long baseY = rect.Height() / 2;
#endif // D_MFC

	const long v = 64;

	long x[3], y[3];
	long idx[3];
	x[0] = bgnX;
	x[1] = bgnX;
	x[2] = bgnX;
	y[0] = baseY - (TITLE_FONT_DOT * 120 / _100_PERCENT);
	y[1] = baseY;
	y[2] = baseY + (TITLE_FONT_DOT * 120 / _100_PERCENT);
	for( long j = 0; j < 3; j++ )
		idx[j] = 0;

	for( long f = 0; f < frame; f++ ){
		if( (gStrTitle[0][idx[0]] == '\0')
				&& (gStrTitle[1][idx[1]] == '\0')
				&& (gStrTitle[2][idx[2]] == '\0') ){
			break;
		}

		for( long j = 0; j < 3; j++ ){
			x[j] -= v;
			if( x[j] < (baseX + (idx[j] * TITLE_FONT_DOT)) ){
				x[j] = (baseX + (idx[j] * TITLE_FONT_DOT));
				char c;
				c = gStrTitle[j][idx[j]];
				if( c != '\0' ){
					drawTitleCharStd( x[j], y[j], c );
					idx[j]++;
					x[j] = bgnX;
				}
			}
		}
	}

	for( long j = 0; j < 3; j++ ){
		if( gStrTitle[j][idx[j]] != '\0' ){
			drawTitleCharStd( x[j], y[j],
					gStrTitle[j][idx[j]] );
		}
	}
}

void PcgDun::drawTitle1( long frame )
{
	if( pTitleStyle == NULL )
		initTitle();

	long len = 1;
	for( long i = 0; i < 3; i++ )
		len = max_l( len, str_len_draw( gStrTitle[i] ) );

#ifdef D_GTK
	const long w = gMapDrawingArea->allocation.width;
	const long h = gMapDrawingArea->allocation.height;
#endif // D_GTK

#ifdef D_MFC
	CXlnlView *view = theApp.m_pMainFrm->GetView();
	CRect rect;
	view->GetClientRect( &rect );

	const long w = rect.Width();
	const long h = rect.Height();
#endif // D_MFC

	const long baseX = (w - (len * TITLE_FONT_DOT)) / 2;
	const long baseY = h / 2;

	long mx[3][16], my[3][16];
	long x[3][16], y[3][16];
	for( long j = 0; j < len; j++ ){
		mx[0][j] = (baseX + (j * TITLE_FONT_DOT));
		mx[1][j] = (baseX + (j * TITLE_FONT_DOT));
		mx[2][j] = (baseX + (j * TITLE_FONT_DOT));
		my[0][j] = baseY - (TITLE_FONT_DOT * 120 / _100_PERCENT);
		my[1][j] = baseY;
		my[2][j] = baseY + (TITLE_FONT_DOT * 120 / _100_PERCENT);

		long xx = (w / 2)
				+ ((len / 2 - j) * TITLE_FONT_DOT);
		x[0][j] = xx;
		x[1][j] = xx;
		x[2][j] = xx;
		y[0][j] = 0 - TITLE_FONT_DOT;
		y[1][j] = h / 2;
		y[2][j] = h + TITLE_FONT_DOT;
	}

	for( long f = 0; f < frame; f++ ){
		for( long i = 0; i < 3; i++ ){
			for( long j = 0; j < len; j++ ){
				x[i][j] += (mx[i][j] - x[i][j]) / 16;
				y[i][j] += (my[i][j] - y[i][j]) / 16;
				x[i][j] += sgn_l( mx[i][j] - x[i][j] );
				y[i][j] += sgn_l( my[i][j] - y[i][j] );
			}
		}
	}

	for( long i = 0; i < 3; i++ ){
		for( long j = 0; j < len; j++ ){
			if( gStrTitle[i][j] == '\0' )
				break;

			drawTitleCharStd( x[i][j], y[i][j],
					gStrTitle[i][j] );
		}
	}
}

void PcgDun::drawTitleCharStd( long x, long y, char c )
{
#ifdef D_GTK
	GdkColor color;

	// 

	color.red = 0x4444;
	color.green = 0x4444;
	color.blue = 0x4444;
	drawTitleChar( color, x + 2, y + 2, c );

	color.red = 0xaaaa;
	color.green = 0x8888;
	color.blue = 0x8888;
	drawTitleChar( color, x - 2, y - 2, c );

	// 

	color.red = 0x8888;
	color.green = 0x0000;
	color.blue = 0x0000;
	drawTitleChar( color, x, y, c );
#endif // D_GTK

#ifdef D_MFC
	D3DCOLOR color;

	// 

	color = 0xff444444;
	drawTitleChar( color, x + 2, y + 2, c );

	color = 0xffaa8888;
	drawTitleChar( color, x - 2, y - 2, c );

	// 

	color = 0xff880000;
	drawTitleChar( color, x, y, c );
#endif // D_MFC
}

#ifdef D_GTK
void PcgDun::drawTitleChar( GdkColor color, long x, long y, char c )
{
	char str[4 + 1];
	str[0] = c;
	str[1] = '\0';

	GdkDrawable *d = getWBuf()->getPixMap();
	GdkFont *font = gtk_style_get_font( pTitleStyle );
	GdkGC *gc = pTitleStyle->fg_gc[GTK_STATE_NORMAL];

	gdk_color_alloc( gdk_colormap_get_system(), &color );
	gdk_gc_set_foreground( gc, &color );

	gdk_draw_string( d, font, gc, x, y, str );
}
#endif // D_GTK

#ifdef D_MFC
void PcgDun::drawTitleChar( D3DCOLOR color, long x, long y, char c )
{
	char str[4 + 1];
	str[0] = c;
	str[1] = '\0';

	mD3Title.Begin();
	mD3Title.Draw( x, y, color, str );
	mD3Title.End();
}
#endif // D_MFC

////////////////////////////////////////////////////////////////
// 饯
// long mapX : X ɸ
// long mapY : Y ɸ
// bool flgFlush : ե졼ࡦХåե饦ɥ
//     褹뤫
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawTile( long mapX, long mapY, bool flgFlush )
{
	if( g_flg_text_mode )
		return drawText( mapX, mapY, 1, 1 );

	setUpdateFlg( mapX, mapY, true );

	if( chk_nest_flg_dun() )
		return true;
	if( !clipWin( mapX, mapY ) )
		return true;

	if( !drawObj( mapX, mapY ) )
		return false;
	if( !drawTrap( mapX, mapY ) )
		return false;
	if( !drawStairsUp( mapX, mapY ) )
		return false;
	if( !drawStairsDown( mapX, mapY ) )
		return false;
	if( !drawSquare( mapX, mapY ) )
		return false;
	if( !drawItemAll( mapX, mapY ) )
		return false;
	if( !drawChrList( mapX, mapY, true ) )
		return false;
	if( !drawCrsrAll( mapX, mapY ) )
		return false;

	if( flgFlush )
		flush( mapX, mapY, 1, 1 );

	setUpdateFlg( mapX, mapY, false );

	return true;
}

////////////////////////////////////////////////////////////////
// ֥Ȥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawObj( long mapX, long mapY )
{
	bool flgSuccess = false;

	if( !clip_pos( mapX, mapY ) )
		return true;

	if( !getWBuf()->getReady() )
		return false;
#ifdef D_WS
	if( !MapForm->getdev()->getReady() )
		return false;
#endif // D_WS

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( TRUE, mapX, mapY, &mjr, &mnr );

	switch( mjr ){
	default:
	case FACE_MJR_NULL:
		flgSuccess = drawNull( mapX, mapY );
		break;
	case FACE_MJR_FLOOR:
		flgSuccess = drawFloor( mapX, mapY );
		break;
	case FACE_MJR_WALL:
		flgSuccess = drawWall( mapX, mapY );
		if( ('A' <= mnr) && (mnr <= 'Z') )
			flgSuccess = drawSignboard( mapX, mapY );
		break;
	case FACE_MJR_DOOR_CLOSE:
		flgSuccess = drawWall( mapX, mapY );
		flgSuccess = drawDoorClose( mapX, mapY );
		break;
	case FACE_MJR_DOOR_OPEN:
		flgSuccess = drawWall( mapX, mapY );
		flgSuccess = drawDoorOpen( mapX, mapY );
		break;
	case FACE_MJR_STAIRS_UP:
	case FACE_MJR_STAIRS_DOWN:
	case FACE_MJR_TRAP:
		flgSuccess = drawFloor( mapX, mapY );
		break;
	case FACE_MJR_WEAPON:
	case FACE_MJR_ARMOR:
	case FACE_MJR_POTION:
	case FACE_MJR_SCROLL:
	case FACE_MJR_STICK:
	case FACE_MJR_ACCESSORY:
	case FACE_MJR_FOOD:
	case FACE_MJR_CHEST:
	case FACE_MJR_COIN:
	case FACE_MJR_GEM:
	case FACE_MJR_MISC:
		flgSuccess = drawFloor( mapX, mapY );
		break;
	}

	return flgSuccess;
}

////////////////////////////////////////////////////////////////
// Υƥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawItemAll( long mapX, long mapY )
{
	bool flgSuccess = false;

	if( !clip_pos( mapX, mapY ) )
		return true;

	if( !getWBuf()->getReady() )
		return false;
#ifdef D_WS
	if( !MapForm->getdev()->getReady() )
		return false;
#endif // D_WS

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( TRUE, mapX, mapY, &mjr, &mnr );

	switch( mjr ){
	default:
	case FACE_MJR_NULL:
		flgSuccess = true;
		break;
	case FACE_MJR_WEAPON:
	case FACE_MJR_ARMOR:
	case FACE_MJR_POTION:
	case FACE_MJR_SCROLL:
	case FACE_MJR_STICK:
	case FACE_MJR_ACCESSORY:
	case FACE_MJR_FOOD:
	case FACE_MJR_CHEST:
	case FACE_MJR_COIN:
	case FACE_MJR_GEM:
	case FACE_MJR_MISC:
		flgSuccess = drawItem( mapX, mapY, mjr, mnr );
		break;
	}

	return flgSuccess;
}

////////////////////////////////////////////////////////////////
// 
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawNull( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long n = aaMapNull[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aNullTown[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	} else {
		return aNullDun[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// 
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawFloor( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long n = aaMapFloor[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aFloorTown[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	} else {
		return aFloorDun[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// ɤ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawWall( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long bit = 0;

	// Τդȹ礻

	if( chkDrawWall( mapX + 0, mapY - 1 ) )
		bit |= gBitIdxWallN;

	if( chkDrawWall( mapX + 0, mapY + 1 ) )
		bit |= gBitIdxWallS;

	if( chkDrawWall( mapX + 1, mapY + 0 ) )
		bit |= gBitIdxWallE;

	if( chkDrawWall( mapX - 1, mapY + 0 ) )
		bit |= gBitIdxWallW;

	long idxX = gIdxWall[bit].x;
	long idxY = gIdxWall[bit].y;

	// ƤդäƤ뤫?

	long bitAll = (gBitIdxWallN | gBitIdxWallS
			| gBitIdxWallE | gBitIdxWallW);
	if( (bit & bitAll) == bitAll ){
		// ФΤդȹ礻

		long bitDiagonal = 0;

		if( chkDrawWall( mapX - 1, mapY - 1 ) )
			bitDiagonal |= gBitIdxWallNW;

		if( chkDrawWall( mapX + 1, mapY - 1 ) )
			bitDiagonal |= gBitIdxWallNE;

		if( chkDrawWall( mapX - 1, mapY + 1 ) )
			bitDiagonal |= gBitIdxWallSW;

		if( chkDrawWall( mapX + 1, mapY + 1 ) )
			bitDiagonal |= gBitIdxWallSE;

		idxX = gIdxWallDiagonal[bitDiagonal].x;
		idxY = gIdxWallDiagonal[bitDiagonal].y;
	}

	long n = aaMapWall[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aWallTown[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX,
				idxY,
				getTileSizeX(),
				getTileSizeY() );
	} else {
		return aWallDun[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX,
				idxY,
				getTileSizeX(),
				getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// Ĥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawSignboard( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;
	if( !clip_draw_pos( mapX, mapY ) )
		return true;

	// Ĥü?

	long dx = 0;
	if( !chkDrawSignboard( mapX - 1, mapY ) )
		dx--;
	if( !chkDrawSignboard( mapX + 1, mapY ) )
		dx++;
	long idxX = dx + 1;

	long n = aaMapSignboard[mapY][mapX];

	bool result = aSignboard[n][0].drawIdx(
			getWBuf(),
			mapX * getTileSizeX(),
			mapY * getTileSizeY(),
			1 * getTileSizeX(),
			1 * getTileSizeY(),
			idxX,
			0,
			getTileSizeX(),
			getTileSizeY() );
	if( !result )
		return result;

	dun_t *dun = get_dun();

#ifdef D_GTK
	GdkDrawable *d = getWBuf()->getPixMap();
	GdkFont *font = gtk_style_get_font( pSignboardStyle );
	GdkGC *gc = pSignboardStyle->fg_gc[GTK_STATE_NORMAL];

	// طʿ

	GdkColor color;

	color.red = 0x0000;
	color.green = 0x0000;
	color.blue = 0x0000;
	gdk_color_alloc( gdk_colormap_get_system(), &color );
	gdk_gc_set_background( gc, &color );

	// ɸ׻

	long x = mapX * getTileSizeX();
	long y = mapY * getTileSizeY();
	x += (nSignboardIntervalWidth / 2) + nSignboardLeftBearing;
	y += (nSignboardIntervalHeight / 2) + nSignboardAscent;

	// 

	char s[4 + 1];
	s[0] = dun->map.obj.mnr[mapY][mapX];
	s[1] = '\0';

	// 

	color.red = 0x4444;
	color.green = 0x4444;
	color.blue = 0x4444;
	gdk_color_alloc( gdk_colormap_get_system(), &color );
	gdk_gc_set_foreground( gc, &color );

	gdk_draw_string( d, font, gc, x + 2, y + 2, s );

	// 

	color.red = 0xffff;
	color.green = 0xffff;
	color.blue = 0xffff;
	gdk_color_alloc( gdk_colormap_get_system(), &color );
	gdk_gc_set_foreground( gc, &color );

	gdk_draw_string( d, font, gc, x, y, s );

	//

	return true;
#endif // D_GTK

#ifdef D_MFC
	// طʿ

#if FLAG_USE_TEXT_OUT
	CXlnlView *view = theApp.m_pMainFrm->GetView();
	CClientDC dc( view );

	COLORREF color;
	color = RGB( 0x0000, 0x0000, 0x0000 );
	dc.SetBkColor( color );
#else
	D3DCOLOR d3color;
#endif

	// ɸ׻

	long scrollX = getScrollBarX();
	long scrollY = getScrollBarY();
	long sizX = getTileSizeX( true );
	long sizY = getTileSizeY( true );
	long offsetX = scrollX / sizX;
	long offsetY = scrollY / sizY;
	long x = (mapX * sizX) - scrollX;
	long y = (mapY * sizY) - scrollY;
	x += nSignboardIntervalWidth / 2;
	y += nSignboardIntervalHeight / 2;
	x = x * g_Dir3d.nTileSizeRate / _100_PERCENT;
	y = y * g_Dir3d.nTileSizeRate / _100_PERCENT;

	// 

	char s[4 + 1];
	s[0] = dun->map.obj.mnr[mapY][mapX];
	s[1] = '\0';

	if( pSignboardFont == NULL )
		return false;

	// 

#if FLAG_USE_TEXT_OUT
	CFont *pPrevFont = dc.SelectObject( pSignboardFont );

	color = RGB( 0x4444, 0x4444, 0x4444 );
	dc.SetTextColor( color );
	dc.TextOut( x + 2, y + 2, s );
#else
	mD3Signboard.Begin();

	d3color = 0xff444444;
	mD3Signboard.Draw( x + 2, y + 2, d3color, s );
#endif

	// 

#if FLAG_USE_TEXT_OUT
	color = RGB( 0xffff, 0xffff, 0xffff );
	dc.SetTextColor( color );
	dc.TextOut( x, y, s );

	dc.SelectObject( pPrevFont );
#else
	d3color = 0xffffffff;
	mD3Signboard.Draw( x, y, d3color, s );

	mD3Signboard.End();
#endif

	return true;
#endif // D_MFC

	return false;
}

////////////////////////////////////////////////////////////////
// 곬ʤ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawStairsUp( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( TRUE, mapX, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_STAIRS_UP )
		return true;

	long n = aaMapStairsUp[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aStairsUpTown[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	} else {
		return aStairsUpDun[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// 곬ʤ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawStairsDown( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( TRUE, mapX, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_STAIRS_DOWN )
		return true;

	long n = aaMapStairsDown[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aStairsDownTown[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	} else {
		return aStairsDownDun[n][0].drawFix( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// Ĥɥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawDoorClose( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long idxX = +1;
	long idxY = +1;
	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;

	set_face_from_map_obj( FALSE, mapX - 1, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_CLOSE )
		idxX += -1;
	set_face_from_map_obj( FALSE, mapX + 1, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_CLOSE )
		idxX += +1;

	set_face_from_map_obj( FALSE, mapX, mapY - 1, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_CLOSE )
		idxY += -1;
	set_face_from_map_obj( FALSE, mapX, mapY + 1, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_CLOSE )
		idxY += +1;

	long n = aaMapDoorClose[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aDoorCloseTown[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX, idxY,
				getTileSizeX(), getTileSizeY() );
	} else {
		return aDoorCloseDun[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX, idxY,
				getTileSizeX(), getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// ɥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawDoorOpen( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long idxX = +1;
	long idxY = +1;
	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;

	set_face_from_map_obj( FALSE, mapX - 1, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_OPEN )
		idxX += -1;
	set_face_from_map_obj( FALSE, mapX + 1, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_OPEN )
		idxX += +1;

	set_face_from_map_obj( FALSE, mapX, mapY - 1, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_OPEN )
		idxY += -1;
	set_face_from_map_obj( FALSE, mapX, mapY + 1, &mjr, &mnr );
	if( mjr != FACE_MJR_DOOR_OPEN )
		idxY += +1;

	long n = aaMapDoorOpen[mapY][mapX];

	if( get_dun()->lev == 0 ){
		return aDoorOpenTown[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX, idxY,
				getTileSizeX(), getTileSizeY() );
	} else {
		return aDoorOpenDun[n][0].drawIdx( getWBuf(),
				mapX * getTileSizeX(),
				mapY * getTileSizeY(),
				1 * getTileSizeX(),
				1 * getTileSizeY(),
				idxX, idxY,
				getTileSizeX(), getTileSizeY() );
	}

	return true;
}

////////////////////////////////////////////////////////////////
// 櫤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawTrap( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	trap_t *trap = get_trap( mapX, mapY );
	if( trap == NULL )
		return true;

	trap_tab_t *tab = get_trap_tab( trap->kind );
	if( tab == NULL )
		return false;

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( TRUE, mapX, mapY, &mjr, &mnr );
	if( mjr != FACE_MJR_TRAP )
		return true;

	long n = 0;
	for( long i = 0; i < PCG_TRAP_MAX_N; i++ ){
		if( gPcgTabTrap[i].nRatio <= -1 ){
			n = 0;
			break;
		}
		if( gPcgTabTrap[i].mnr == tab->mnr ){
			n = i;
			break;
		}
	}
	if( !trap->flg_chked )
		n = 0;

	if( trap->flg_chked && (trap->kind == TRAP_KIND_SANCTUARY) )
		return drawTrapSanctuary( mapX, mapY, trap, tab, n );

	long x = getObjDrawLeft( mapX, &(aTrap[n][0]) );
	long y = getObjDrawTop( mapY, &(aTrap[n][0]) );
	long w = getObjDrawWidth( mapX, &(aTrap[n][0]) );
	long h = getObjDrawHeight( mapX, &(aTrap[n][0]) );

	return( aTrap[n][0].draw( getWBuf(), x, y, w, h ) );
}

////////////////////////////////////////////////////////////////
// 
// long mapX : X ɸ
// long mapY : Y ɸ
// trap_t *trap : 
// trap_tab_t *tab : Υơ֥
// long nPcg : ѥΥХꥨ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawTrapSanctuary(
	long mapX, long mapY,
	trap_t *trap, trap_tab_t *tab,
	long nPcg
)
{
	if( !clip_pos( mapX, mapY ) )
		return true;
	if( trap == NULL )
		return false;
	if( tab == NULL )
		return false;
	if( trap->kind != TRAP_KIND_SANCTUARY )
		return false;

	mapX += trap->cx;
	mapY += trap->cy;

	long x = getObjDrawLeft( mapX, &(aTrap[nPcg][0]) );
	long y = getObjDrawTop( mapY, &(aTrap[nPcg][0]) );
	long w = getObjDrawWidth( mapX, &(aTrap[nPcg][0]) );
	long h = getObjDrawHeight( mapX, &(aTrap[nPcg][0]) );

	return( aTrap[nPcg][0].draw( getWBuf(), x, y, w, h ) );
}

////////////////////////////////////////////////////////////////
// ᥤȥ֤Υ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawCrsrAll( long mapX, long mapY )
{
	bool flagNoErr = true;

	if( !drawCrsrSub( mapX, mapY ) )
		flagNoErr = false;

	if( !drawCrsr( mapX, mapY ) )
		flagNoErr = false;

	return flagNoErr;
}

////////////////////////////////////////////////////////////////
// 
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawCrsr( long mapX, long mapY )
{
	if( !clipWin( mapX, mapY ) )
		return true;

	pos_t *crsr = get_crsr();
	if( crsr == NULL )
		return true;
	if( !clip_pos( crsr->x, crsr->y ) )
		return true;
	if( !clipWin( crsr->x, crsr->y ) )
		return true;

	if( !chkDrawCrsr( mapX, mapY ) )
		return true;

	long crsrW = aCrsr[0][0].getWidth();
	long crsrH = aCrsr[0][0].getHeight();
	long crsrX = (crsr->x * getTileSizeX());
	long crsrY = (crsr->y * getTileSizeY());
	crsrX += (getTileSizeX() / 2) - (crsrW / 2);
	crsrY += (getTileSizeY() / 2) - (crsrH / 2);

	long drawW = getTileSizeX();
	long drawH = getTileSizeY();
	long drawX = mapX * getTileSizeX();
	long drawY = mapY * getTileSizeY();

	long w = (drawX + drawW) - (crsrX + crsrW);
	long h = (drawY + drawH) - (crsrY + crsrH);
	if( w > 0 )
		drawW -= w;
	if( h > 0 )
		drawH -= h;
	if( drawW < 1 )
		return true;
	if( drawH < 1 )
		return true;

	long offsetX = drawX - crsrX;
	long offsetY = drawY - crsrY;

	return( aCrsr[0][0].drawOffset( getWBuf(),
			drawX, drawY, drawW, drawH, offsetX, offsetY ) );
}

////////////////////////////////////////////////////////////////
// ֡
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawCrsrSub( long mapX, long mapY )
{
	if( !clipWin( mapX, mapY ) )
		return true;

	pos_t *crsrSub = get_sub_crsr();
	if( crsrSub == NULL )
		return true;
	if( !clip_pos( crsrSub->x, crsrSub->y ) )
		return true;
	if( !clipWin( crsrSub->x, crsrSub->y ) )
		return true;

	if( !chkDrawCrsrSub( mapX, mapY ) )
		return true;

	long crsrSubW = aCrsrSub[0][0].getWidth();
	long crsrSubH = aCrsrSub[0][0].getHeight();
	long crsrSubX = (crsrSub->x * getTileSizeX());
	long crsrSubY = (crsrSub->y * getTileSizeY());
	crsrSubX += (getTileSizeX() / 2) - (crsrSubW / 2);
	crsrSubY += (getTileSizeY() / 2) - (crsrSubH / 2);

	long drawW = getTileSizeX();
	long drawH = getTileSizeY();
	long drawX = mapX * getTileSizeX();
	long drawY = mapY * getTileSizeY();

	long w = (drawX + drawW) - (crsrSubX + crsrSubW);
	long h = (drawY + drawH) - (crsrSubY + crsrSubH);
	if( w > 0 )
		drawW -= w;
	if( h > 0 )
		drawH -= h;
	if( drawW < 1 )
		return true;
	if( drawH < 1 )
		return true;

	long offsetX = drawX - crsrSubX;
	long offsetY = drawY - crsrSubY;

	return( aCrsrSub[0][0].drawOffset( getWBuf(),
			drawX, drawY, drawW, drawH, offsetX, offsetY ) );
}

////////////////////////////////////////////////////////////////
// 󥫡
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawSquare( long mapX, long mapY )
{
	if( !clipWin( mapX, mapY ) )
		return true;

	square_t *square = get_square_mbr();
	if( square == NULL )
		return false;

	if( !clip_pos( square->x, square->y ) )
		return true;
	if( !clipWin( square->x, square->y ) )
		return true;
	if( !chkDrawSquare( mapX, mapY ) )
		return true;

	if( nSquareN[square->dir] <= 0 )
		return false;

	long squareW = aSquare[square->dir][0].getWidth();
	long squareH = aSquare[square->dir][0].getHeight();
	long squareX = (square->x * getTileSizeX());
	long squareY = (square->y * getTileSizeY());
	squareX += (getTileSizeX() / 2) - (squareW / 2);
	squareY += (getTileSizeY() / 2) - (squareH / 2);

	long drawW = getTileSizeX();
	long drawH = getTileSizeY();
	long drawX = mapX * getTileSizeX();
	long drawY = mapY * getTileSizeY();

	long w = (drawX + drawW) - (squareX + squareW);
	long h = (drawY + drawH) - (squareY + squareH);
	if( w > 0 )
		drawW -= w;
	if( h > 0 )
		drawH -= h;
	if( drawW < 1 )
		return true;
	if( drawH < 1 )
		return true;

	long offsetX = drawX - squareX;
	long offsetY = drawY - squareY;

	return( aSquare[square->dir][0].drawOffset( getWBuf(),
			drawX, drawY, drawW, drawH, offsetX, offsetY ) );
}

////////////////////////////////////////////////////////////////
// ƥ
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawItem( long mapX, long mapY, char mjr, char mnr )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	long n = 0;
	for( long i = 0; i < LOOP_MAX_1000; i++ ){
		if( gPcgTabItem[i].nRatio <= -1 ){
			n = 0;
			break;
		}
		if( (gPcgTabItem[i].mjr == mjr)
				&& (gPcgTabItem[i].mnr == mnr) ){
			n = i;
			break;
		}
	}

	long x = getObjDrawLeft( mapX, &(aItem[n][0]) );
	long y = getObjDrawTop( mapY, &(aItem[n][0]) );
	long w = getObjDrawWidth( mapX, &(aItem[n][0]) );
	long h = getObjDrawHeight( mapX, &(aItem[n][0]) );

	return( aItem[n][0].draw( getWBuf(), x, y, w, h ) );
}

////////////////////////////////////////////////////////////////
// 饯
// long x : ̤ X ɸ
// long y : ̤ Y ɸ
// long w : ̤
// long h : ̤ι⤵
////////////////////////////////////////////////////////////////

void PcgDun::drawChrListAll( long x, long y, long w, long h )
{
	chr_t *chr = get_sort_chr_head();

	for( long i = 0; i < LOOP_MAX_10000;
			i++, chr = chr->sort_chr_next ){
		if( chr == NULL )
			break;

		Pcg *pPcg = getChrPcg( chr );

		long x1 = getChrDrawLeft( chr, pPcg, false );
		if( x1 >= x + w )
			continue;

		long y1 = getChrDrawTop( chr, pPcg, false );
		if( y1 >= y + h )
			continue;

		long x2 = getChrDrawRight( chr, pPcg, false );
		if( x2 < x )
			continue;

		long y2 = getChrDrawBottom( chr, pPcg, false );
		if( y2 < y )
			continue;

		drawChr( chr->x, chr->y, chr, pPcg, false );
	}
}

////////////////////////////////////////////////////////////////
// ɸϰ饯
// long mapX : X ɸ
// long mapY : Y ɸ
// bool flagOneTile : 1 ʬ褹뤫
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawChrList( long mapX, long mapY, bool flagOneTile )
{
	if( !clipWin( mapX, mapY ) )
		return true;

	chr_t *chr = get_sort_chr_head();

	for( long i = 0; i < LOOP_MAX_10000;
			i++, chr = chr->sort_chr_next ){
		if( chr == NULL )
			break;

		if( !chkDrawChr( mapX, mapY, chr ) )
			continue;

		Pcg *pPcg = getChrPcg( chr );
		if( !drawChr( mapX, mapY, chr, pPcg, flagOneTile ) ){
			return false;
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////
// ɸϰ饯
// long mapX : X ɸ
// long mapY : Y ɸ
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// bool flagOneTile : 1 ʬ褹뤫
// return : 顼̵ä
////////////////////////////////////////////////////////////////

bool PcgDun::drawChr(
	long mapX, long mapY,
	chr_t *chr, Pcg *pPcg,
	bool flagOneTile
)
{
	if( chr == NULL )
		return false;
	if( pPcg == NULL )
		return false;

	if( flagOneTile && !clipWin( mapX, mapY ) )
		return true;

	long x1 = getChrDrawLeft( chr, pPcg, false );
	long y1 = getChrDrawTop( chr, pPcg, false );
	long x2 = getChrDrawRight( chr, pPcg, false );
	long y2 = getChrDrawBottom( chr, pPcg, false );

	long xx1 = x1;
	long yy1 = y1;
	long xx2 = x2;
	long yy2 = y2;

	if( flagOneTile ){
		if( xx1 < ((mapX + 0) * getTileSizeX()) )
			xx1 = ((mapX + 0) * getTileSizeX());
		if( yy1 < ((mapY + 0) * getTileSizeY()) )
			yy1 = ((mapY + 0) * getTileSizeY());

		if( xx2 > ((mapX + 1) * getTileSizeX()) )
			xx2 = ((mapX + 1) * getTileSizeX());
		if( yy2 > ((mapY + 1) * getTileSizeY()) )
			yy2 = ((mapY + 1) * getTileSizeY());
	}

	long w = xx2 - xx1;
	long h = yy2 - yy1;
	if( w < 1 )
		return true;
	if( h < 1 )
		return true;

	long offsetX = xx1 - x1;
	long offsetY = yy1 - y1;

	bool flgDir4 = false;
	if( (chr->kind == CHR_KIND_MBR)
			&& (nPtnDirMbr == N_PTN_DIR_CHR_4) ){
		flgDir4 = true;
	}
	if( (chr->kind == CHR_KIND_MNSTR)
			&& (nPtnDirMnstr == N_PTN_DIR_CHR_4) ){
		flgDir4 = true;
	}
	if( flgDir4 ){
		long dirIdx = get_chr_dir_idx( chr );
		dirIdx *= getHeightChr( chr, pPcg );
		offsetY += dirIdx;
	}

	return( pPcg->drawOffset( getWBuf(),
			xx1, yy1, w, h, offsetX, offsetY ) );
}

////////////////////////////////////////////////////////////////
// 饯Υѥ֤
// chr_t *chr : 饯
// return : ѥ
////////////////////////////////////////////////////////////////

Pcg *PcgDun::getChrPcg( chr_t *chr )
{
	if( chr == NULL )
		return NULL;

	if( chr->kind == CHR_KIND_MBR ){
		return( &(aMbr[chr->ls_mbr_n]) );
	} else if( chr->kind == CHR_KIND_MNSTR ){
		long n = chr->mnstr_kind;
		long m = 0;

		if( nMnstrN[n] <= 0 )
			n = MNSTR_KIND_NULL;

		return( &(aMnstr[n][m]) );
	}

	return NULL;
}

////////////////////////////////////////////////////////////////
// 饯κüκɸ֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// bool flagPre : 1 κɸ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getChrDrawLeft( chr_t *chr, Pcg *pPcg, bool flagPre )
{
	if( chr == NULL )
		return MAP_DEL_X;
	if( pPcg == NULL )
		return MAP_DEL_X;

	long bx = 0;
	long ex = 0;
	if( flagPre ){
		bx = get_chr_left( chr->pre_draw_x, chr->dx );
		ex = (get_chr_right( chr->pre_draw_x, chr->dx ) + 1);
	} else {
		bx = get_chr_left( chr->x, chr->dx );
		ex = (get_chr_right( chr->x, chr->dx ) + 1);
	}
	bx *= getTileSizeX();
	ex *= getTileSizeX();

	long x1 = ((ex + bx) / 2) - (getWidthChr( chr, pPcg ) / 2);

	return x1;
}

////////////////////////////////////////////////////////////////
// 饯αüκɸ֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// bool flagPre : 1 κɸ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getChrDrawRight( chr_t *chr, Pcg *pPcg, bool flagPre )
{
	if( chr == NULL )
		return MAP_DEL_X;
	if( pPcg == NULL )
		return MAP_DEL_X;

	long bx = 0;
	long ex = 0;
	if( flagPre ){
		bx = get_chr_left( chr->pre_draw_x, chr->dx );
		ex = (get_chr_right( chr->pre_draw_x, chr->dx ) + 1);
	} else {
		bx = get_chr_left( chr->x, chr->dx );
		ex = (get_chr_right( chr->x, chr->dx ) + 1);
	}
	bx *= getTileSizeX();
	ex *= getTileSizeX();

	long x2 = ((ex + bx) / 2) + (getWidthChr( chr, pPcg ) / 2);

	return x2;
}

////////////////////////////////////////////////////////////////
// 饯ξüκɸ֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// bool flagPre : 1 κɸ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getChrDrawTop( chr_t *chr, Pcg *pPcg, bool flagPre )
{
	if( chr == NULL )
		return MAP_DEL_Y;
	if( pPcg == NULL )
		return MAP_DEL_Y;

	long ey = 0;
	if( flagPre ){
		ey = (get_chr_bottom( chr->pre_draw_y, chr->dy ) + 1);
		ey = (get_chr_bottom( chr->pre_draw_y, chr->dy ) + 1);
	} else {
		ey = (get_chr_bottom( chr->y, chr->dy ) + 1);
		ey = (get_chr_bottom( chr->y, chr->dy ) + 1);
	}
	ey *= getTileSizeY();

	long y1 = ey - getHeightChr( chr, pPcg );

	return y1;
}

////////////////////////////////////////////////////////////////
// 饯βüκɸ֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// bool flagPre : 1 κɸ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getChrDrawBottom( chr_t *chr, Pcg *pPcg, bool flagPre )
{
	if( chr == NULL )
		return MAP_DEL_Y;
	if( pPcg == NULL )
		return MAP_DEL_Y;

	long ey = 0;
	if( flagPre ){
		ey = (get_chr_bottom( chr->pre_draw_y, chr->dy ) + 1);
		ey = (get_chr_bottom( chr->pre_draw_y, chr->dy ) + 1);
	} else {
		ey = (get_chr_bottom( chr->y, chr->dy ) + 1);
		ey = (get_chr_bottom( chr->y, chr->dy ) + 1);
	}
	ey *= getTileSizeY();

	long y2 = ey;

	return y2;
}

////////////////////////////////////////////////////////////////
// ֥Ȥκüκɸ֤
// long mapX : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawLeft( long mapX, Pcg *pPcg )
{
	long tileSize = getTileSizeX();

	long w = pPcg->getWidth();
	long x = mapX * tileSize;
	long dx = (tileSize - w) / 2;

	return( x + dx );
}

////////////////////////////////////////////////////////////////
// ֥Ȥαüκɸ֤
// long mapX : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawRight( long mapX, Pcg *pPcg )
{
	long left = getObjDrawLeft( mapX, pPcg );

	return( left + pPcg->getWidth() );
}

////////////////////////////////////////////////////////////////
// ֥Ȥ֤
// long mapX : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : 
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawWidth( long mapX, Pcg *pPcg )
{
	return( pPcg->getWidth() );
}

////////////////////////////////////////////////////////////////
// ֥Ȥξüκɸ֤
// long mapY : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawTop( long mapY, Pcg *pPcg )
{
	long tileSize = getTileSizeY();

	long h = pPcg->getHeight();
	long y = mapY * tileSize;
	long dy = (tileSize - h) / 2;

	return( y + dy );
}

////////////////////////////////////////////////////////////////
// ֥Ȥβüκɸ֤
// long mapY : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : ɸ
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawBottom( long mapY, Pcg *pPcg )
{
	long top = getObjDrawTop( mapY, pPcg );

	return( top + pPcg->getHeight() );
}

////////////////////////////////////////////////////////////////
// ֥Ȥι⤵֤
// long mapY : ޥåפκɸ
// Pcg *pPcg : ѥ
// return : ⤵
////////////////////////////////////////////////////////////////

long PcgDun::getObjDrawHeight( long mapY, Pcg *pPcg )
{
	return( pPcg->getHeight() );
}

////////////////////////////////////////////////////////////////
// κɸɤĴ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : ɤ
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawWall( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	char mjr = FACE_MJR_NULL;
	set_face_from_map_obj( FALSE, mapX, mapY, &mjr, NULL );

	switch( mjr ){
	case FACE_MJR_WALL:
	case FACE_MJR_DOOR_OPEN:
	case FACE_MJR_DOOR_CLOSE:
		return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////
// κɸĤĴ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : Ĥ
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawSignboard( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return true;

	char mjr = FACE_MJR_NULL;
	char mnr = FACE_MNR_NULL;
	set_face_from_map_obj( FALSE, mapX, mapY, &mjr, &mnr );

	if( mjr != FACE_MJR_WALL )
		return false;
	if( mnr < 'A' )
		return false;
	if( mnr > 'Z' )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// κɸ褬뤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 褬뤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawSanctuary( long mapX, long mapY, trap_t *trap )
{
	if( trap == NULL )
		return false;
	if( !clip_pos( mapX, mapY ) )
		return false;
	if( !clip_pos( trap->x, trap->y ) )
		return false;

	long x = mapX * getTileSizeX();
	long y = mapY * getTileSizeY();
	long w = aTrap[0][0].getWidth();
	long h = aTrap[0][0].getHeight();

	long bx = (trap->x * getTileSizeX()) - (w / 2);
	long by = (trap->y * getTileSizeY()) - (h / 2);
	long ex = bx + w;
	long ey = by + h;

	if( x < bx )
		return false;
	if( y < by )
		return false;
	if( x > ex )
		return false;
	if( y > ey )
		return false;

	return true;
}
////////////////////////////////////////////////////////////////
// κɸ˥뤬뤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 뤬뤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawCrsr( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return false;

	pos_t *crsr = get_crsr();
	if( crsr == NULL )
		return false;
	if( !clip_pos( crsr->x, crsr->y ) )
		return false;

	long x = mapX * getTileSizeX();
	long y = mapY * getTileSizeY();
	long w = aCrsr[0][0].getWidth();
	long h = aCrsr[0][0].getHeight();

	long bx = (crsr->x * getTileSizeX()) - (w / 2);
	long by = (crsr->y * getTileSizeY()) - (h / 2);
	long ex = bx + w;
	long ey = by + h;

	if( x < bx )
		return false;
	if( y < by )
		return false;
	if( x > ex )
		return false;
	if( y > ey )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// κɸ˥֡뤬뤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 뤬뤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawCrsrSub( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return false;

	pos_t *crsrSub = get_sub_crsr();
	if( crsrSub == NULL )
		return false;
	if( !clip_pos( crsrSub->x, crsrSub->y ) )
		return false;

	long x = mapX * getTileSizeX();
	long y = mapY * getTileSizeY();
	long w = aCrsrSub[0][0].getWidth();
	long h = aCrsrSub[0][0].getHeight();

	long bx = (crsrSub->x * getTileSizeX()) - (w / 2);
	long by = (crsrSub->y * getTileSizeY()) - (h / 2);
	long ex = bx + w;
	long ey = by + h;

	if( x < bx )
		return false;
	if( y < by )
		return false;
	if( x > ex )
		return false;
	if( y > ey )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// κɸ˥󥫡뤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// return : 󥫡뤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawSquare( long mapX, long mapY )
{
	if( !clip_pos( mapX, mapY ) )
		return false;

	square_t *square = get_square_mbr();
	if( square == NULL )
		return false;
	if( !clip_pos( square->x, square->y ) )
		return false;

	long x = mapX * getTileSizeX();
	long y = mapY * getTileSizeY();
	long w = aSquare[square->dir][0].getWidth();
	long h = aSquare[square->dir][0].getHeight();

	long bx = (square->x * getTileSizeX()) - (w / 2);
	long by = (square->y * getTileSizeY()) - (h / 2);
	long ex = bx + w;
	long ey = by + h;

	if( x < bx )
		return false;
	if( y < by )
		return false;
	if( x > ex )
		return false;
	if( y > ey )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// 饯Υѥ󤬻κɸϰ⤫Ĵ٤
// long mapX : X ɸ
// long mapY : Y ɸ
// chr_t *chr : 饯
// return : ϰ⤫
////////////////////////////////////////////////////////////////

bool PcgDun::chkDrawChr( long mapX, long mapY, chr_t *chr )
{
	if( chr == NULL )
		return false;

	if( !clip_pos( mapX, mapY ) )
		return false;
	if( !clipWin( mapX, mapY ) )
		return false;
	if( chr->kind == CHR_KIND_MNSTR )
		if( !chk_flg( chr->flg_map, FLG_MAP_CHR_FIND ) )
			return false;
	if( chk_flg_or( chr->stat, FLG_STAT_NOT_EXIST | FLG_STAT_DEAD ) )
		return false;

	Pcg *pPcg = getChrPcg( chr );

	long x1 = getChrDrawLeft( chr, pPcg, false );
	long y1 = getChrDrawTop( chr, pPcg, false );
	long x2 = getChrDrawRight( chr, pPcg, false );
	long y2 = getChrDrawBottom( chr, pPcg, false );

	long sx = getTileSizeX();
	long sy = getTileSizeY();

	if( x1 >= ((mapX + 1) * sx) )
		return false;
	if( y1 >= ((mapY + 1) * sy) )
		return false;

	if( x2 < ((mapX + 0) * sx) )
		return false;
	if( y2 < ((mapY + 0) * sy) )
		return false;

	return true;
}

////////////////////////////////////////////////////////////////
// 1 饯ʬΥեåѥ֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// return : 
////////////////////////////////////////////////////////////////

long PcgDun::getWidthChr( chr_t *chr, Pcg *pPcg )
{
	if( chr == NULL )
		return 0;
	if( pPcg == NULL )
		return 0;

	long n = 1;
	if( chr->kind == CHR_KIND_MBR )
		n = nPtnDirMbr;
	if( chr->kind == CHR_KIND_MNSTR )
		n = nPtnDirMnstr;

	return( pPcg->getWidth() / n );
}

////////////////////////////////////////////////////////////////
// 1 饯ʬΥեåѥι⤵֤
// chr_t *chr : 饯
// Pcg *pPcg : ѥ
// return : ⤵
////////////////////////////////////////////////////////////////

long PcgDun::getHeightChr( chr_t *chr, Pcg *pPcg )
{
	if( chr == NULL )
		return 0;
	if( pPcg == NULL )
		return 0;

	long n = 1;
	if( chr->kind == CHR_KIND_MBR )
		n = nPtnDirMbr;
	if( chr->kind == CHR_KIND_MNSTR )
		n = nPtnDirMnstr;

	return( pPcg->getHeight() / n );
}
